from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Optional, Union, Iterable, TypeVar, Tuple

from ...util import tracing
from ...util.error import err, exc, Source, Part
from ...util.naming import sanitize
from ...util.structures import OrderedSet

from . import metamodel as mm, builtins as bt
from .rewriter import Walker, Rewriter, NO_WALK
from .builtins import builtins as b

#--------------------------------------------------
# Typer
#--------------------------------------------------

T = TypeVar("T", bound=mm.Model | mm.Task)

class Typer:
    def __init__(self, enforce=False):
        self.enforce = enforce
        self.model_net: Optional[PropagationNetwork] = None
        # TODO: remove this once we are using spans for the debugger
        self.last_net: Optional[PropagationNetwork] = None

    def infer_model(self, model: mm.Model) -> mm.Model:
        """
        Infer types for the given model, returning a new model with updated types and
        storing the results within this typer so that subsequent calls can reuse them.
        """
        # create a brand new network with data from this model
        self.model_net = PropagationNetwork(model)
        return self._infer(model, self.model_net)

    def infer_query(self, query: mm.Task) -> mm.Task:
        """
        Infer types for the given query, returning a new query with updated types. If the
        typer was used to infer a model previously, the information from that model analysis
        will be reused.
        """
        if self.model_net is not None:
            net = PropagationNetwork(self.model_net.model)
            net.load_types(self.model_net.resolved_types)
        else:
            net = PropagationNetwork(mm.Model(root=query))
        self.last_net = net
        return self._infer(query, net)

    #--------------------------------------------------
    # Internal implementation
    #--------------------------------------------------

    def _infer(self, node: T, net: PropagationNetwork) -> T:
        # build the propagation network by analyzing the model
        with tracing.span("typer.analyze"):
            Analyzer(net).analyze(node)

        # propagate the types through the network
        with tracing.span("typer.propagate") as span:
            net.propagate()
            # span["type_graph"] = net.to_mermaid()

        # replace the fields in the model with the new types
        with tracing.span("typer.replace"):
            replacer = Replacer(net)
            final = replacer.rewrite(node)

        # report any errors found during typing
        if net.errors:
            for error in net.errors:
                error.report()
            if self.enforce:
                exc("TyperError", "Type errors detected during type inference.")
        return final


#--------------------------------------------------
# Propagation Network
#--------------------------------------------------

# The core idea of the typer is to build a propagation network that represents data
# dependencies between nodes. Nodes can be variables, literals, fields and tasks that refer
# to relations (lookups, aggregates, updates). An edge from node A to node B means that
# in order to resolve references and types for node B we need to know the type of node A.
#
# After building the network, we start with roots that have known types (literals, fields
# loaded from previous analysis and vars with concrete types) and propagate their types via
# the edges to other nodes. It is possible that the network contains cycles, so we iterate
# on the work list until fixpoint is reached, i.e. when we cannot propagate any new type.
#
# During the propagation, we are only gathering information about the types and references
# of nodes. Once we reach a fixpoint, we do a final pass to rewrite the model with the new
# type information, which may include adding casts to convert between types where needed.
#
Node = Union[mm.Var, mm.Field, mm.Literal, mm.Lookup, mm.Aggregate, mm.Update]

class PropagationNetwork():
    def __init__(self, model: mm.Model):
        # the model that is either being analyzed or that is the context for a query
        self.model = model

        # the resolved types of nodes in the network
        self.resolved_types: dict[Node, mm.Type] = {}

        # map from unresolved placeholder relations to their potential target replacements
        self.potential_targets: dict[mm.Relation, list[mm.Relation]] = {}

        # nodes loaded from previous analysis, which we use as roots to start propagation
        self.loaded_roots = set()

        # edges in the propagation network, from one node to potentially many
        self.edges:dict[Node, OrderedSet[Node]] = defaultdict(lambda: OrderedSet())
        self.back_edges:dict[Node, OrderedSet[Node]] = defaultdict(lambda: OrderedSet())

        # all errors collected during inference
        self.errors:list[TyperError] = []

        # overloads resolved for a lookup/update/aggregate, by node id. This is only for
        # relations that declare overloads
        self.resolved_overload:dict[int, mm.Overload] = {}

        # placeholders resolved for a lookup, by node id. This is only for relations that
        # are placeholders (i.e. only Any fields) and will be replaced by references to
        # these concrete relations. E.g. a query for "name(Any, Any)" may be replaced by
        # the union of "name(Dog, String)" and name(Cat, String)".
        self.resolved_placeholder:dict[int, list[mm.Relation]] = {}

        # for a given lookup/update/aggregate that involves numbers, the specific number
        # type resolved for it.
        self.resolved_number:dict[int, mm.NumberType] = {}

    #--------------------------------------------------
    # Error reporting
    #--------------------------------------------------

    def type_mismatch(self, node: Node|mm.Update, expected: mm.Type, actual: mm.Type):
        self.errors.append(TypeMismatch(node, expected, actual))

    def invalid_type(self, node: Node, type: mm.Type):
        self.errors.append(InvalidType(node, type))

    def unresolved_overload(self, node: mm.Lookup|mm.Aggregate):
        # TODO - consider renaming this to UnresolvedReference
        self.errors.append(UnresolvedOverload(node, [self.resolve(a) for a in node.args]))

    def unresolved_type(self, node: Node):
        # TODO - this is not being used yet, we need a pass at the end to check for any node
        # that we could not resolve
        self.errors.append(UnresolvedType(node))

    def has_errors(self, node: Node) -> bool:
        for mismatch in self.errors:
            if mismatch.node == node:
                return True
        return False

    #--------------------------------------------------
    # Types and Edges
    #--------------------------------------------------

    def add_edge(self, source: Node, target: Node):
        self.edges[source].add(target)
        self.back_edges[target].add(source)

    def add_resolved_type(self, node: Node, type: mm.Type):
        """ Register that this node was resolved to have this type. """
        if node in self.resolved_types:
            self.resolved_types[node] = merge_types(self.resolved_types[node], type)
        else:
            self.resolved_types[node] = type

    #--------------------------------------------------
    # Load types from a previous analysis
    #--------------------------------------------------

    def load_types(self, type_dict: dict[Node, mm.Type]):
        for node, type in type_dict.items():
            if isinstance(node, (mm.Field)):
                self.add_resolved_type(node, type)
                self.loaded_roots.add(node)

    #--------------------------------------------------
    # Resolve Values
    #--------------------------------------------------

    def resolve(self, value: mm.Value) -> mm.Type:
        if isinstance(value, (mm.Var, mm.Literal)):
            return self.resolved_types.get(value) or to_type(value)
        if isinstance(value, mm.Field):
            return self.resolved_types.get(value) or value.type
        assert not isinstance(value, (mm.Task)), "Should never try to resolve a task"
        return to_type(value)

    #--------------------------------------------------
    # Resolve References
    #--------------------------------------------------

    def all_dependencies_resolved(self, op:mm.Lookup|mm.Aggregate):
        """ True iff all dependencies required to resolve this reference are met. """
        rel = get_relation(op)
        # if this is a placeholder, eq or cast, we assume all possible args were resolved
        if bt.is_placeholder(rel) or rel == b.core.eq or rel == b.core.cast:
            return True

        # else, find whether all back-edges were resolved
        for node in self.back_edges[op]:
            # cannot resolve if a required var, literal or input field is still abstract
            if isinstance(node, (mm.Var, mm.Literal)) or (isinstance(node, mm.Field) and node.input):
                node_type = self.resolve(node)
                if bt.is_abstract(node_type):
                    return False
        return True

    def resolve_reference(self, op: mm.Lookup|mm.Aggregate) -> Optional[mm.Overload|list[mm.Relation]]:
        # check if all dependencies required to resolve this reference are met
        if not self.all_dependencies_resolved(op):
            return None

        relation = get_relation(op)
        if bt.is_placeholder(relation):
            # when replacing a placeholder, we may have multiple matches
            matches = []
            resolved_args = [self.resolve(arg) for arg in op.args]
            for target in self.potential_targets.get(relation, []):
                fields = get_relation_fields(target, relation.name)
                if all(type_matches(arg, self.resolve(field))
                        for arg, field in zip(resolved_args, fields)):
                    matches.append(target)
            return matches

        elif relation.overloads:
            # when resolving an overload for a concrete relation, we can only have one
            resolved_args = [self.resolve(arg) for arg in op.args]
            is_function = bt.is_function(relation)
            for overload in relation.overloads:
                # note that for functions we only consider input fields when matching
                if all(type_matches(arg, field_type) or conversion_allowed(arg, field_type)
                       for arg, field, field_type in zip(resolved_args, relation.fields, overload.types)
                       if field.input or not is_function):
                    self.resolved_overload[op.id] = overload
                    return overload
            return []  # no matches found

        else:
            # this is a relation with type vars or numbers that needs to be specialized
            return [relation]

    #--------------------------------------------------
    # Propagation
    #--------------------------------------------------

    def propagate(self):
        edges = self.edges
        work_list = []

        # start with the loaded roots + all literals + sources of edges without back edges
        work_list.extend(self.loaded_roots)
        for source in self.edges.keys():
            if isinstance(source, (mm.Literal)) or not source in self.back_edges:
                work_list.append(source)

        # limit the number of iterations to avoid infinite loops
        i = 0
        max_iterations = 100 * len(self.edges)

        # propagate types until we reach a fixed point
        while work_list:
            i += 1
            if i > max_iterations:
                err("Infinite Loop", "Infinite loop detected in the typer. Please, report this as a bug.")
                break

            node = work_list.pop(0)
            next = None
            if isinstance(node, mm.Field):
                # this is a field loaded from a previous analysis, so all we need to do is
                # propagate its type to its output edges
                next = edges.get(node, [])

            elif isinstance(node, (mm.Lookup, mm.Aggregate)):
                # this is a lookup/aggregate that may be overloaded or a placeholder; try
                # to resolve its reference, i.e. determine which specific relation or
                # overload it refers to
                found = self.resolve_reference(node)
                if found is not None:
                    # if found is None it means that we need more info to resolve the
                    # reference; otherwise it was possible to resolve it (even if to no matches)

                    # keep the resolved args before propagation to see if they will change
                    resolved_args = [self.resolve(arg) for arg in node.args]
                    # propagate the reference resolution
                    self.propagate_reference(node, resolved_args, found)
                    if found:
                        # the next nodes to process are all the outgoing edges plus any arg that
                        # changed during propagation
                        next = OrderedSet()
                        next.update(edges.get(node, []))
                        next.update([arg for idx, arg in enumerate(node.args) if isinstance(arg, mm.Var) and not (self.resolve(arg) == resolved_args[idx])])
                    else:
                        # the reference is unresolved, so remove from the worklist the args
                        # because they depend on this being resolved
                        for arg in node.args:
                            if arg in work_list:
                                work_list.remove(arg)
            else:
                assert isinstance(node, (mm.Var, mm.Literal))
                resolved = self.resolve(node)
                # if we ended up with an invalid type, report it on a next edge (which is
                # where the var or literal is going to be used)
                if invalid_type(resolved):
                    self.invalid_type(edges.get(node, [node])[0], resolved)

                # if the var is still abstract, add it to the work list to try again later
                if isinstance(node, mm.Var) and bt.is_abstract(resolved):
                    # but if everything that is left are vars, we cannot make progress
                    if not all(isinstance(x, mm.Var) for x in work_list):
                        next = [node]
                else:
                    # otherwise, check all outgoing edges
                    next = []
                    for out in edges.get(node, []):
                        if isinstance(out, mm.Update):
                            # this is an update to a field. We have to check that the type is
                            # valid for the field and if the field is abstract we can refine the
                            # type. But if the field is concrete we have to check that the type
                            # matche or can be converted.
                            is_population_lookup = isinstance(out.relation, mm.TypeNode)
                            for field in get_update_fields(node, out):
                                if not type_matches(resolved, field.type, accept_expected_super_types=is_population_lookup) and not conversion_allowed(resolved, field.type):
                                    self.type_mismatch(out, field.type, resolved)
                                elif bt.is_abstract(field.type):
                                    # if the field type changed, propagate further
                                    if resolved != self.resolve(field):
                                        next.append(field)
                                    self.add_resolved_type(field, resolved)
                        else:
                            # this is an arg flowing into a task, so resolve the task next
                            next.append(out)

            # add the new nodes to the work list
            if next is not None:
                for n in next:
                    if n not in work_list:
                        work_list.append(n)


    def propagate_reference(self, task:mm.Lookup|mm.Aggregate, resolved_args:list[mm.Type], references:mm.Overload|list[mm.Relation]):
        # TODO: distinguish between overloads and placeholders better when raising errors
        if not references:
            return self.unresolved_overload(task)

        relation = get_relation(task)

        if relation == b.core.cast:
            # cast is special cased: we just propagate the type to the target
            cast_type, target = task.args[0], task.args[2]
            assert isinstance(cast_type, mm.Type)
            assert isinstance(target, mm.Var)
            self.add_resolved_type(target, cast_type)
            return

        if bt.is_placeholder(relation):
            assert(references and isinstance(references, list))
            # we've resolved the placeholder, so store that
            self.resolved_placeholder[task.id] = references

            # we need to determine the final types of our args by taking all the references
            # and adding the type of their fields back to the args.
            for reference in references:
                resolved_fields = [self.resolve(f) for f in get_relation_fields(reference, relation.name)]
                for field_type, arg_type, arg in zip(resolved_fields, resolved_args, task.args):
                    if bt.is_abstract(arg_type) and isinstance(arg, mm.Var):
                        self.add_resolved_type(arg, field_type)
        else:
            if isinstance(references, mm.Overload):
                # we resolved an overload, use its types
                types = list(references.types)
            else:
                # we resolved to a single concrete relation that contains type vars or needs
                # number specialization, so use that relation's field types
                types = list([self.resolve(f) for f in relation.fields])

            # if our overload preserves types, we check to see if there's a preserved output
            # type given the inputs and if so, shadow the field type with the preserved type.
            # this is only attempted if all input types match the field types, i.e. no
            # conversions are needed
            resolved_fields = types
            if bt.is_function(relation) and len(set(resolved_fields)) == 1 and not relation in self.NON_TYPE_PRESERVERS and\
                all(type_matches(arg_type, field_type) for arg_type, field_type, field in zip(resolved_args, types, relation.fields) if field.input):

                input_types = set([arg_type for field, arg_type
                                    in zip(relation.fields, resolved_args) if field.input])
                if out_type := self.try_preserve_type(input_types):
                    resolved_fields = [field_type if field.input else out_type
                                        for field, field_type in zip(relation.fields, types)]


            # eq is special cased because we don't want to specialize a number for it, as it
            # can be just comparing numbers of different types.
            if relation != b.core.eq and (b.core.Number in types or (b.core.TypeVar in types and any(bt.is_number(t) for t in resolved_args))):
                # this relation contains generic numbers or typevars bound to numbers, so
                # find which specific type of number to use given the arguments being passed
                number, resolved_fields = self.specialize_number(relation, resolved_fields, resolved_args)
                self.resolved_number[task.id] = number
                # push the specialized number to the outputs
                for field, field_type, arg in zip(relation.fields, resolved_fields, task.args):
                    if not field.input and isinstance(arg, mm.Var):
                        self.add_resolved_type(arg, field_type)

            elif b.core.TypeVar in types:
                # this relation contains type vars, so we have to make sure that all args
                # bound to the same type var are consistent

                # find which arg is bound to the type var and check that they are all consistent
                typevar_type = None
                for arg_type, field_type in zip(resolved_args, types):
                    if field_type == b.core.TypeVar:
                        if typevar_type is None:
                            typevar_type = arg_type
                        else:
                            typevar_type = merge_types(typevar_type, arg_type)
                assert typevar_type is not None

                # compute the final arg types by replacing type vars with the typevar_type
                computed_arg_types = []
                for arg_type, field_type in zip(resolved_args, types):
                    # check that the arg type matches the type var type
                    if not type_matches(typevar_type, arg_type) and not conversion_allowed(arg_type, typevar_type):
                        self.type_mismatch(task, typevar_type, arg_type)
                    if field_type == b.core.TypeVar:
                        computed_arg_types.append(typevar_type)
                    else:
                        computed_arg_types.append(arg_type)

                # if no mismatches were found, propagate the computed arg types back to the args
                if len(computed_arg_types) == len(task.args):
                    for computed_type, arg, arg_type in zip(computed_arg_types, task.args, resolved_args):
                        # TODO: we could allow for non-numeric/string literals because this means
                        # the literal is being used as a value type, but the backend emitters
                        # would have to deal with that.
                        if isinstance(arg, mm.Var) or (isinstance(arg, mm.Literal) and (bt.is_numeric(computed_type) or computed_type == b.core.String)):
                            self.add_resolved_type(arg, computed_type)

            else:
                # no typevar or number specialization shenanigans, just propagate field types to args
                for field, field_type, arg, arg_type in zip(relation.fields, resolved_fields, task.args, resolved_args):
                    # add a resolved type if we learned more about the arg's type
                    if isinstance(arg, mm.Var) and (bt.is_abstract(arg_type) or bt.extends(field_type, arg_type)):
                        self.add_resolved_type(arg, field_type)
                    elif not type_matches(arg_type, field_type) and not conversion_allowed(arg_type, field_type):
                        self.type_mismatch(task, field_type, arg_type)


    # we try to preserve types for relations that are functions (i.e. potentially multiple
    # input but a single output) and where all types are the same. However, there are some
    # exceptions to this rule, e.g. range() always returns int regardless of whether the
    # input type is an int or a value type that extends int.
    NON_TYPE_PRESERVERS = [
        b.common.range
    ]

    def try_preserve_type(self, types:set[mm.Type]) -> Optional[mm.Type]:
        # we keep the input type as the output type if either all inputs are the exact same
        # type or there's one nominal and its base primitive type, e.g. USD + Decimal
        if len(types) == 1:
            return next(iter(types))
        if len(types) == 2:
            t1, t2 = types
            t1_base = bt.get_primitive_supertype(t1)
            t2_base = bt.get_primitive_supertype(t2)
            if t1_base is None or t2_base is None:
                base_equivalent = type_matches(t1, t2, accept_expected_super_types=True)
            else:
                base_equivalent = type_matches(t1_base, t2_base)
            if base_equivalent:
                # as long as one of the types is a base primitive, we can use the
                # other type as final preserved type
                if bt.is_primitive(t1):
                    return t2
                elif bt.is_primitive(t2):
                    return t1
        return None

    def specialize_number(self, op, field_types:list[mm.Type], arg_types:list[mm.Type]) -> Tuple[mm.NumberType, list[mm.Type]]:
        """
        Find the number type to use for an overload that has Number in its field_types, and
        which is being referred to with these arg_types.

        Return a tuple where the first element is the specialized number type, and the second
        element is a new list that contains the same types as field_types but with
        Number replaced by this specialized number.
        """
        # special case a few operators according to Snowflake's rules in
        # https://docs.snowflake.com/en/sql-reference/operators-arithmetic#scale-and-precision-in-arithmetic-operations
        if op == b.core.div:
            # see https://docs.snowflake.com/en/sql-reference/operators-arithmetic#division
            numerator, denominator = get_number_type(arg_types[0]), get_number_type(arg_types[1])
            s = max(numerator.scale, min(numerator.scale + 6, 12))
            number = mm.NumberType(name=f"Number(38,{s})", precision=38, scale=s, source=op.source, super_types=(b.core.Numeric,))
            return number, [numerator, denominator, number]
        elif op == b.core.mul:
            # see https://docs.snowflake.com/en/sql-reference/operators-arithmetic#multiplication
            t1, t2 = get_number_type(arg_types[0]), get_number_type(arg_types[1])
            S1 = t1.scale
            S2 = t2.scale
            s = min(S1 + S2, max(S1, S2, 12))
            number = mm.NumberType(name=f"Number(38,{s})", precision=38, scale=s, source=op.source, super_types=(b.core.Numeric,))
            return number, [t1, t2, number]
        elif op == b.aggregates.avg:
            # TODO!! - implement proper avg specialization
            pass

        # fall back to the the current specialization policy, which is to select the number
        # with largest scale and, if there multiple with the largest scale, the one with the
        # largest precision. This is safe because when converting a number to the
        # specialized number, we never truncate fractional digits (because we selected the
        # largest scale) and, if the non-fractional digits are too large to fit the
        # specialized number, we will have a runtime overflow, which should alert the user
        # of the problem.
        number = None
        for arg_type in arg_types:
            x = bt.get_number_supertype(arg_type)
            if isinstance(x, mm.NumberType):
                if number is None or x.scale > number.scale or (x.scale == number.scale and x.precision > number.precision):
                    number = x
        assert(number is not None)
        return number, [number if bt.is_number(t) else t for t in field_types]


    #--------------------------------------------------
    # Display
    #--------------------------------------------------

    # draw the network as a mermaid graph for the debugger
    def to_mermaid(self, max_edges=500) -> str:
        # add links for edges while collecting nodes
        nodes = OrderedSet()
        link_strs = []
        # edges
        for src, dsts in self.edges.items():
            nodes.add(src)
            for dst in dsts:
                if len(link_strs) > max_edges:
                    break
                nodes.add(dst)
                link_strs.append(f"n{src.id} --> n{dst.id}")
            if len(link_strs) > max_edges:
                break

        def type_span(t:mm.Type) -> str:
            type_str = t.name if isinstance(t, mm.ScalarType) else str(t)
            return f"<span style='color:cyan;'>{type_str.strip()}</span>"

        def reference_span(rel:mm.Relation, arg_types:list[mm.Type]) -> str:
            args = []
            for field, arg_type in zip(rel.fields, arg_types):
                field_type = self.resolve(field)
                if not type_matches(arg_type, field_type) and not conversion_allowed(arg_type, field_type) and not bt.is_abstract(field_type):
                    args.append(f"<span style='color:yellow;'>{str(arg_type).strip()} -> {str(field_type).strip()}</span>")
                elif isinstance(arg_type, mm.UnionType):
                    args.append(type_span(field_type))
                else:
                    args.append(type_span(arg_type))
            return f'{rel.name}({", ".join(args)})'

        resolved = self.resolved_types
        node_strs = []
        for node in nodes:
            klass = ""
            if isinstance(node, mm.Var):
                ir_type = resolved.get(node) or self.resolve(node)
                type_str = type_span(ir_type)
                label = f'(["{node.name}:{type_str}"])'
            elif isinstance(node, mm.Literal):
                ir_type = resolved.get(node) or self.resolve(node)
                type_str = type_span(ir_type)
                klass = ":::literal"
                label = f'(["{node.value}: {type_str}"])'
            elif isinstance(node, mm.Field):
                ir_type = resolved.get(node) or self.resolve(node)
                type_str = type_span(ir_type)
                klass = ":::field"
                rel = node._relation
                if rel is not None:
                    rel = str(node._relation)
                    label = f'[/"{node.name}:{type_str}\nfrom {rel}"\\]'
                else:
                    label = f'[/"{node.name}:\n{type_str}"\\]'
            elif isinstance(node, (mm.Lookup, mm.Aggregate, mm.Update)):
                arg_types = [self.resolve(arg) for arg in node.args]
                if node.id in self.resolved_placeholder:
                    overloads = self.resolved_placeholder[node.id]
                    content = "<br/>".join([reference_span(o, arg_types) for o in overloads])
                else:
                    content = reference_span(get_relation(node), arg_types)
                if isinstance(node, mm.Update):
                    klass = ":::update"
                    label = f'{{{{"{content}"}}}}'
                else:
                    klass = ":::reference"
                    label = f'[/"{content}"/]'
            else:
                raise NotImplementedError(f"Unknown node type: {type(node)}")
            if self.has_errors(node):
                klass = ":::error"
            node_strs.append(f'n{node.id}{label}{klass}')

        node_str = "\n                ".join(node_strs)
        link_str = "\n                ".join(link_strs)
        template = f"""
            %%{{init: {{'theme':'dark', 'flowchart':{{'useMaxWidth':false, 'htmlLabels': true}}}}}}%%
            flowchart TD
                linkStyle default stroke:#666
                classDef field fill:#245,stroke:#478
                classDef update fill:#245,stroke:#478
                classDef literal fill:#452,stroke:#784
                classDef error fill:#624,stroke:#945,color:#f9a
                classDef default stroke:#444,stroke-width:2px, font-size:12px

                %% nodes
                {node_str}

                %% edges
                {link_str}
        """
        return template

#--------------------------------------------------
# Analyzer
#--------------------------------------------------

class Analyzer(Walker):
    """ Walks the metamodel and builds the propagation network. """

    def __init__(self, net:PropagationNetwork):
        super().__init__()
        self.net = net

    def analyze(self, node: mm.Node):
        self(node)

    # TODO - ignoring requires for now because the typing of constraints seems incorrect
    def enter_require(self, require: mm.Require):
        return NO_WALK

    def compute_potential_targets(self, relation: mm.Relation):
        # register potential targets for placeholders
        if bt.is_placeholder(relation):
            self.net.potential_targets[relation] = get_potential_targets(self.net.model, relation)

    #--------------------------------------------------
    # Walk Update
    #--------------------------------------------------

    def update(self, node: mm.Update):
        rel = node.relation
        self.compute_potential_targets(rel)

        # arg is flowing into a field
        for arg, field in zip(node.args, rel.fields):
            if isinstance(arg, (mm.Var, mm.Literal)):
                self.net.add_edge(arg, node)

    #--------------------------------------------------
    # Walk Lookups + Aggregates
    #--------------------------------------------------

    def lookup(self, task: mm.Lookup):
        self.compute_potential_targets(task.relation)
        self.visit_rel_op(task)

    def aggregate(self, task: mm.Aggregate):
        self.visit_rel_op(task)

    def visit_rel_op(self, task: mm.Lookup|mm.Aggregate):
        relation = get_relation(task)

        is_placeholder = bt.is_placeholder(relation)
        for field, arg in zip(relation.fields, task.args):
            if isinstance(arg, (mm.Var, mm.Literal)):
                if field.input:
                    # we need to resolve all inputs before resolving the relation
                    self.net.add_edge(arg, task)
                else:
                    # placeholders also need the output to be resolved
                    if is_placeholder:
                        self.net.add_edge(arg, task)
                    else:
                        # args bound to outputs can be resolved after
                        self.net.add_edge(task, arg)
                        # if the field is abstract, it needs to be resolved before we can
                        # resolve the task.
                        if bt.is_abstract(self.net.resolve(field)):
                            self.net.add_edge(field, task)


#--------------------------------------------------
# Replacer
#--------------------------------------------------

# Once we've pushed all the types through the network, we need to replace the types of
# fields and vars that we may have discovered. We also need to replace placeholder lookups
# with the chosen relations and do any conversions that are needed.
class Replacer(Rewriter):
    def __init__(self, net:PropagationNetwork):
        super().__init__()
        self.net = net

    def rewrite(self, model: T) -> T:
        return self(model) # type: ignore

    def logical(self, logical: mm.Logical):
        if len(logical.body) == 0:
            return logical
        # inline logicals that are just there to group other nodes during rewrite
        body = []
        for child in logical.body:
            if isinstance(child, mm.Logical) and not child.optional and not child.scope:
                body.extend(child.body)
            else:
                body.append(child)
        return logical.mut(body = tuple(body))

    #--------------------------------------------------
    # Rewriter handlers
    #--------------------------------------------------

    def field(self, node: mm.Field):
        # TODO - this is only modifying the relation in the model, but then we have a new
        # relation there, which is different than the object referenced by tasks.
        if node in self.net.resolved_types:
            return mm.Field(node.name, self.net.resolved_types[node], node.input, _relation = node._relation)
        return node

    def var(self, node: mm.Var):
        if node in self.net.resolved_types:
            return mm.Var(self.net.resolved_types[node], node.name)
        return node

    def literal(self, node: mm.Literal):
        if node in self.net.resolved_types:
            return mm.Literal(self.net.resolved_types[node], node.value)
        return node

    def update(self, node: mm.Update):
        return self.convert_arguments(node, node.relation)

    def lookup(self, node: mm.Lookup):
        # We need to handle eq specially because its arguments can be converted symmetrically
        if node.relation == b.core.eq:
            return self.visit_eq_lookup(node)

        args = types = None
        if node.id in self.net.resolved_placeholder:
            # placeholder resolved to multiple relations
            resolved_relations = self.net.resolved_placeholder[node.id]
            args = get_lookup_args(node, resolved_relations[0])
            types = [f.type for f in resolved_relations[0].fields]
        elif node.id in self.net.resolved_overload:
            # overload resolved to a specific relation
            resolved_relations = [node.relation]
            types = self.net.resolved_overload[node.id].types
        else:
            # single relation
            resolved_relations = [node.relation]

        if len(resolved_relations) == 1:
            # single relation, just convert arguments
            x = self.convert_arguments(node, resolved_relations[0], args, types)
            if isinstance(x, mm.Logical) and len(x.body) == 1:
                return x.body[0]
            else:
                return x

        # multiple relations, create a union
        branches:list = []
        for target in resolved_relations:
            args = get_lookup_args(node, target)
            types = [f.type for f in get_relation_fields(resolved_relations[0], node.relation.name)]
            # adding this logical to avoid issues in the old backend
            branches.append(mm.Logical((self.convert_arguments(node, target, args, types=types, force_copy=True),)))
        return mm.Union(tuple(branches))

    def convert_arguments(self, node: mm.Lookup|mm.Update, relation: mm.Relation, args: Iterable[mm.Value]|None=None, types: Iterable[mm.Type]|None=None, force_copy=False) -> mm.Logical|mm.Lookup|mm.Update:
        """ This node was resolved to target this relation using these args, which should
        have these types. Convert any arguments as needed and return a new node with the
        proper relation and converted args. If multiple conversions are needed, return a
        logical that contains all the conversion tasks plus the final node. """
        args = args or node.args
        types = types or [self.net.resolve(f) for f in relation.fields]
        number_type = self.net.resolved_number.get(node.id)
        is_function = bt.is_function(relation)
        tasks = []
        final_args = []
        for arg, field, field_type in zip(args, relation.fields, types):
            if isinstance(arg, (mm.Var, mm.Literal)) and (not is_function or field.input):
                arg_type = to_type(arg)
                if number_type and bt.is_number(arg_type) and not arg_type == b.core.ScaledNumber:
                    field_type = number_type
                # the typer previously made sure that this should be valid so a type mismatch
                # means we need to convert
                if not type_matches(arg_type, field_type):
                    final_args.append(convert(arg, field_type, tasks))
                else:
                    final_args.append(arg)
            else:
                final_args.append(arg)
        # add the original node with the proper target relation and converted args;
        # here we want to use mut because we keep information about nodes based on their ids
        # (e.g. we store resolved types based on ids), so we want to keep the same id. But
        # when a lookup is being converted as part of a union (i.e. multiple targets for a
        # placeholder), we need to create nodes with new ids to avoid conflicts.
        if force_copy:
            tasks.append(node.replace(relation = relation, args = tuple(final_args)))
        else:
            tasks.append(node.mut(relation = relation, args = tuple(final_args)))
        # if we need conversion tasks, wrap in a logical
        if len(tasks) == 1:
            return tasks[0]
        return mm.Logical(tuple(tasks))

    def visit_eq_lookup(self, node: mm.Lookup):
        (left, right) = node.args
        left_type = to_type(left)
        right_type = to_type(right)

        if type_matches(left_type, right_type):
            return node

        assert isinstance(left, (mm.Var, mm.Literal)) and isinstance(right, (mm.Var, mm.Literal))
        final_args = []
        tasks = []
        if conversion_allowed(left_type, right_type):
            final_args = [convert(left, right_type, tasks), right]
        elif conversion_allowed(right_type, left_type):
            final_args = [left, convert(right, left_type, tasks)]
        else:
            # this type mismatch was reported during propagation, so just return the node
            return node

        tasks.append(mm.Lookup(b.core.eq, tuple(final_args)))
        return mm.Logical(tuple(tasks))

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def get_relation(node: mm.Lookup|mm.Update|mm.Aggregate) -> mm.Relation:
    if isinstance(node, mm.Aggregate):
        return node.aggregation
    return node.relation

def get_name(type: mm.Type) -> str:
    if isinstance(type, mm.ScalarType):
        return type.name
    elif isinstance(type, mm.UnionType):
        return '|'.join([get_name(t) for t in type.types])
    elif isinstance(type, mm.ListType):
        return f'List[{get_name(type.element_type)}]'
    elif isinstance(type, mm.TupleType):
        return f'Tuple[{", ".join([get_name(t) for t in type.element_types])}]'
    else:
        raise TypeError(f"Unknown type: {type}")

#--------------------------------------------------
# Type and Relation helpers
#--------------------------------------------------

def get_update_fields(arg, update: mm.Update) -> Iterable[mm.Field]:
    """ Get the fields of the relation being updated by this arg. Note that an arg can be
    bound to multiple fields at the same time. """
    if arg in update.args:
        for x, field in zip(update.args, update.relation.fields):
            if arg == x:
                yield field
    return []

def get_relation_fields(relation: mm.Relation, name: str) -> Iterable[mm.Field]:
    """ Get the fields of this relation, potentially reordered to match the reading with the given name."""
    if name == relation.name:
        return relation.fields
    for reading in relation.readings:
        if reading.name == name:
            # reorder the fields to match the correct ordering
            fields = []
            for idx in reading.field_order:
                fields.append(relation.fields[idx])
            return fields
    return []

def get_lookup_args(node: mm.Lookup, target: mm.Relation):
    """ Get the args of this lookup, potentially reordered to match the reading with the given name."""
    for reading in target.readings:
        if reading.name == node.relation.name:
            # reorder the args to match the correct ordering
            args = []
            for idx in reading.field_order:
                args.append(node.args[idx])
            return args
    return node.args

def get_number_type(t: mm.Type) -> mm.NumberType:
    # Get a number type from the given type, if it is Number return the default number
    x = bt.get_number_supertype(t)
    if isinstance(x, mm.NumberType):
        return x
    return b.core.DefaultNumber

def is_potential_target(placeholder: mm.Relation, target: mm.Relation) -> bool:
    """ Whether this target is matches the placeholder signature and, thus, can be a potential target. """
    if placeholder != target and len(placeholder.fields) == len(target.fields) and not bt.is_placeholder(target):
        return placeholder.name == target.name or any(placeholder.name == reading.name for reading in target.readings)
    return False

def get_potential_targets(model: mm.Model, placeholder: mm.Relation) -> list[mm.Relation]:
    """ Get all potential target relations in the model that match the placeholder signature. """
    return list(filter(lambda r: is_potential_target(placeholder, r), model.relations))

def to_type(value: mm.Value|mm.Field|mm.Literal) -> mm.Type:
    if isinstance(value, (mm.Var, mm.Literal)):
        return value.type

    if isinstance(value, mm.Type):
        return b.core.Type

    if isinstance(value, mm.Field):
        return b.core.Field

    if isinstance(value, tuple):
        return mm.TupleType(element_types=tuple(to_type(v) for v in value))

    raise TypeError(f"Cannot determine IR type for value: {value} of type {type(value).__name__}")

def convert(value: mm.Var|mm.Literal, to_type: mm.Type, tasks: list[mm.Task]) -> mm.Value:
    # if the arg is a literal, we can just change its type
    # TODO - we may want to check that the value is actually convertible
    if isinstance(value, mm.Literal):
        return mm.Literal(to_type, value.value)

    # otherise we need to add a cast
    name = sanitize(value.name + "_" + get_name(to_type))
    to_type_base = bt.get_primitive_supertype(to_type) or to_type
    new_value = mm.Var(to_type_base, name)
    tasks.append(mm.Lookup(b.core.cast, (to_type_base, value, new_value)))
    return new_value

def conversion_allowed(from_type: mm.Type, to_type: mm.Type) -> bool:
    # value type conversion is allowed only if the value types are related by inheritance
    if bt.is_value_type(from_type) and bt.is_value_type(to_type) and not bt.extends(from_type, to_type):
        return False
    # value type conversion is allowed
    x = bt.get_primitive_supertype(from_type)
    y = bt.get_primitive_supertype(to_type)
    if x and y and (x != from_type or y != to_type) and conversion_allowed(x, y):
        return True

    # numbers can be converted to floats
    if bt.is_numeric(from_type) and to_type == b.core.Float:
        return True

    # a number can be converted to another number of larger scale
    if isinstance(from_type, mm.NumberType) and isinstance(to_type, mm.NumberType):
        if to_type.scale >= from_type.scale:
            return True

    if from_type == b.core.Number and isinstance(to_type, mm.NumberType):
        return True

    return False

def type_matches(actual: mm.Type, expected: mm.Type, accept_expected_super_types=False) -> bool:
    """
    True iff we can use a value of the actual type when expecting the expected type, without
    conversions.

    Any super-type of `actual` can match `expected`. For example if we expect a `Person`, we
    can use an `Employee` if `Employee < Person`.

    In general, the other way around is not true: if we expect an `Employee` we cannot use a
    `Person` instead.

    However, when the relation is a Type (previsously known as "population relations"), it
    is valid to provide sub-types of the expected type. For example, `Employee(Person)` is
    a valid way to check that `Person` is an `Employee` on a `Lookup`, or to assert that a
    particular `Person` is an `Employee` on an `Update`.
    """
    # exact match
    if actual == expected:
        return True

    # any matches anything
    if actual == b.core.Any or expected == b.core.Any:
        return True

    # type vars match anything
    if expected == b.core.TypeVar:
        return True

    # if an entity type var or any entity is expected, it matches any actual entity type
    if (expected == b.core.EntityTypeVar or bt.extends(expected, b.core.AnyEntity)) and not bt.is_primitive(actual):
        return True

    # the abstract Number type and the number type variable match any number type
    if (expected == b.core.Number) and bt.is_number(actual):
        return True

    if (expected == b.core.Numeric) and bt.is_numeric(actual):
        return True

    # different value types never match
    if bt.is_value_type(actual) and bt.is_value_type(expected) and not bt.extends(actual, expected):
        return False

    # if actual is scalar, any of its parents may match the expected type
    if isinstance(actual, mm.ScalarType) and any([type_matches(parent, expected) for parent in actual.super_types]):
        return True

    # if expected is a value type or this is a check for a type relation, any of the expected type's parents may match the actual type
    if (accept_expected_super_types or bt.is_value_type(expected)) and isinstance(expected, mm.ScalarType) and any([type_matches(actual, parent, accept_expected_super_types) for parent in expected.super_types]):
        return True

    # if we expect a union, the actual can match any of its types
    if isinstance(expected, mm.UnionType):
        for t in expected.types:
            if type_matches(t, actual, accept_expected_super_types):
                return True

    # if actual is a union, every one of its types must match the expected type
    # if isinstance(actual, mm.UnionType):
    #     for t in actual.types:
    #         if not type_matches(t, expected, accept_expected_super_types):
    #             return False
    #     return True
    # TODO - we have to distinguish between when we are checking that a specific arg matches
    # a relation vs when we are selecting relations for the placeholders; then we have to
    # decide between the above and this.
    if isinstance(actual, mm.UnionType):
        for t in actual.types:
            if type_matches(t, expected, accept_expected_super_types):
                return True

    # a list type matches if their element types match
    if isinstance(actual, (mm.ListType, mm.TupleType)) and isinstance(expected, mm.ListType):
        if isinstance(actual, mm.TupleType):
            for et in actual.element_types:
                if not type_matches(et, expected.element_type):
                    return False
            return True
        return type_matches(actual.element_type, expected.element_type)

    # a tuple types match if any of all their types match
    if isinstance(actual, mm.TupleType) and isinstance(expected, mm.TupleType):
        return all([type_matches(ae, ee) for ae, ee in zip(actual.element_types, expected.element_types)])

    # accept tuples with a single element type to match a list with that type
    if isinstance(actual, mm.TupleType) and isinstance(expected, mm.ListType):
        return type_matches(actual.element_types[0], expected.element_type)

    # otherwise no match
    return False

def merge_numeric_types(type1: mm.Type, type2: mm.Type) -> Optional[mm.Type]:
    # if one of them is the abstract Number type, pick the other
    if type1 == b.core.Number and isinstance(type2, mm.NumberType):
        return type2
    if type2 == b.core.Number and isinstance(type1, mm.NumberType):
        return type1

    # if both are number types, pick the one with larger scale/precision
    if isinstance(type1, mm.NumberType) and isinstance(type2, mm.NumberType):
        if type1.scale > type2.scale or (type1.scale == type2.scale and type1.precision > type2.precision):
            return type1
        else:
            return type2

    # if we are overriding a number with a float, pick float
    if isinstance(type1, mm.NumberType) and type2 == b.core.Float:
        return type2
    if isinstance(type2, mm.NumberType) and type1 == b.core.Float:
        return type1

    return None

def merge_types(type1: mm.Type, type2: mm.Type) -> mm.Type:
    if type1 == type2:
        return type1
    if bt.is_type_var(type1):
        return type2
    if bt.is_type_var(type2):
        return type1

    types_to_process = [type1, type2]

    numeric_merge = merge_numeric_types(type1, type2)
    if numeric_merge is not None:
        return numeric_merge

    # if one extends the other, pick the most specific one
    if bt.extends(type1, type2):
        return type1
    if bt.extends(type2, type1):
        return type2

    # give precedence to nominal types (e.g. merging USD(decimal) with decimal gives USD(decimal))
    base_primitive_type1 = bt.get_primitive_supertype(type1)
    base_primitive_type2 = bt.get_primitive_supertype(type2)
    if base_primitive_type1 == base_primitive_type2:
        if bt.is_primitive(type1):
            return type2
        elif bt.is_primitive(type2):
            return type1

    if base_primitive_type1 and base_primitive_type2:
        numeric_merge = merge_numeric_types(base_primitive_type1, base_primitive_type2)
        if numeric_merge == base_primitive_type1:
            return type1
        elif numeric_merge == base_primitive_type2:
            return type2

    combined = OrderedSet()
    # Iterative flattening of union types
    while types_to_process:
        t = types_to_process.pop()
        if isinstance(t, mm.UnionType):
            types_to_process.extend(t.types)
        else:
            combined.add(t)

    # If we have multiple types and Any or AnyEntity is one of them, remove Any
    if len(combined) > 1:
        if b.core.Any in combined:
            combined.remove(b.core.Any)
        if b.core.AnyEntity in combined:
            combined.remove(b.core.AnyEntity)

    # If we still have multiple types, make sure supertypes are removed to keep only the
    # most specific types
    if len(combined) > 1:
        to_remove = set()
        for t1 in combined:
            for t2 in combined:
                if t1 != t2 and bt.extends(t1, t2):
                    to_remove.add(t2)
        for r in to_remove:
            combined.remove(r)

    # Return single type or create a union
    return next(iter(combined)) if len(combined) == 1 else mm.UnionType(types=tuple(combined))

def invalid_type(type:mm.Type) -> bool:
    if isinstance(type, mm.UnionType):
        # if there are multiple primitives, or a primitive and a non-primitive
        # then we have an invalid type
        if len(type.types) > 1:
            return any([bt.is_primitive(t) for t in type.types])
    return False


#--------------------------------------------------
# Type Errors
#--------------------------------------------------

@dataclass
class TyperError():
    node: mm.Node

    def report(self):
        err(self.name(), self.message(), self.parts())

    def name(self) -> str:
        return type(self).__name__

    def message(self) -> str:
        raise NotImplementedError()

    def parts(self) -> list[Part]:
        if self.node.source is None:
            return [str(self.node)]
        return [str(self.node), Source(self.node.source)]

@dataclass
class TypeMismatch(TyperError):
    expected: mm.Type
    actual: mm.Type

    def message(self) -> str:
        return f"Expected {get_name(self.expected)}, got {get_name(self.actual)}"

@dataclass
class InvalidType(TyperError):
    type: mm.Type

    def message(self) -> str:
        return f"Incompatible types infered: {get_name(self.type)}"

@dataclass
class UnresolvedOverload(TyperError):
    arg_types: list[mm.Type]

    def message(self) -> str:
        assert isinstance(self.node, (mm.Lookup, mm.Update, mm.Aggregate))
        rel = get_relation(self.node)
        types = ', '.join([get_name(t) for t in self.arg_types])
        return f"Unresolved overload: {rel.name}({types})"

@dataclass
class UnresolvedType(TyperError):

    def message(self) -> str:
        return "Unable to determine concrete type."



# pretty.py

from __future__ import annotations
from typing import Any, Iterable

from .base import *

#------------------------------------------------------
# Utility helpers
#------------------------------------------------------

def _is_multiline(s: str) -> bool:
    return "\n" in s

def _indent(s: str, level: int, unit: str) -> str:
    pad = unit * level
    return "\n".join(pad + line if line else pad for line in s.splitlines())

def _join_inline(parts: Iterable[str]) -> str:
    return ", ".join(parts)

_BUILTIN_LABEL_OVERRIDES: dict[str, str] = {
}

def _relationship_label(rel: Relationship, *, include_owner: bool = True) -> str:
    short = getattr(rel, "_short_name", "")
    if short:
        model = getattr(rel, "_model", None)
        model_name = getattr(model, "name", "")
        if model_name == "Builtins":
            return _BUILTIN_LABEL_OVERRIDES.get(short, short)
        if not include_owner:
            return short
        fields = getattr(rel, "_fields", [])
        if fields:
            owner = fields[0].type
            owner_name = getattr(owner, "_name", None)
            if owner_name:
                return f"{owner_name}.{short}"
        return short
    reading = getattr(rel, "_reading", "")
    if reading:
        return reading
    readings = getattr(rel, "_readings", None)
    if readings:
        return readings[0]._reading
    return ""

def _block(name: str, body_lines: list[str], level: int, unit: str, max_inline_len: int | None = None) -> str:
    """
    Renders:
    (name
      line1
      line2)
    """
    if not body_lines:
        return ""
    if (
        max_inline_len is not None
        and len(body_lines) == 1
        and not _is_multiline(body_lines[0])
    ):
        inline_form = f"({name} {body_lines[0]})"
        if len(inline_form) <= max_inline_len:
            return inline_form
    head = f"({name}"
    inner_lines = [_indent(line, 1, unit) for line in body_lines]
    if inner_lines:
        inner_lines[-1] = f"{inner_lines[-1]})"
    return "\n".join([head] + inner_lines)

#------------------------------------------------------
# The Pretty Printer
#------------------------------------------------------

class PrettyPrinter:
    def __init__(
        self,
        indent_unit: str = "  ",
        max_inline_len: int = 80,
        inline_args_threshold: int = 3,
    ):
        self.indent_unit = indent_unit
        self.max_inline_len = max_inline_len
        self.inline_args_threshold = inline_args_threshold
        # cache already-rendered nodes (protects against cycles / re-visits)
        self._seen = {}

    # Public entrypoint
    def format(self, obj: Any, level: int = 0) -> str:
        key = (id(obj), level)
        if key in self._seen:
            return self._seen[key]

        # dynamic single-dispatch
        meth = None
        for cls in type(obj).__mro__:
            meth = getattr(self, f"_fmt_{cls.__name__}", None)
            if meth:
                break
        if meth is None:
            # Fallback: plain str/repr
            out = str(obj)
            self._seen[key] = out
            return out

        out = meth(obj, level)
        self._seen[key] = out
        return out

    #------------------------------------------------------
    # Leaf-ish nodes
    #------------------------------------------------------

    def _fmt_Concept(self, c: Concept, level: int) -> str:
        return c._name

    def _fmt_Table(self, t: Table, level: int) -> str:
        return t._name  # Table inherits Concept; Chain lives in _name

    def _fmt_Literal(self, lit: Literal, level: int) -> str:
        # lit.__str__ is fine, but we bypass it to avoid nested old-printers
        v = lit._value
        return repr(v)

    def _fmt_Alias(self, a: Alias, level: int) -> str:
        src = self.format(a._source, level)
        return f"{src} AS {a._alias}"

    def _fmt_FieldRef(self, fr: FieldRef, level: int) -> str:
        src = self.format(fr._root, level)
        # render the key as given by the user (str|int|Concept)
        if isinstance(fr._field, (int,)):
            key = f"{fr._field}"
        elif hasattr(fr._field, "_name"):
            key = getattr(fr._field, "_name")
        else:
            key = str(fr._field)
        return f"{src}[{key}]"

    def _fmt_Ref(self, ref: Ref, level: int) -> str:
        concept = getattr(ref, "_concept", None)
        if concept is None:
            return "<ref>"
        concept_label = self.format(concept, level) if isinstance(concept, Variable) else str(concept)
        return f"{concept_label}.ref"

    def _fmt_DerivedColumn(self, col: DerivedColumn, level: int) -> str:
        if col._name:
            return col._name  # user-provided alias

        table = col._table
        index = col._index

        if table is not None and isinstance(index, int):
            if isinstance(table, Fragment):
                source_values = table._select
                if isinstance(source_values, (list, tuple)) and 0 <= index < len(source_values):
                    return self.format(source_values[index], level)
            table_label = table.__class__.__name__
            return f"{table_label}[{index}]"

        return f"col{index}" if index is not None else "<col>"

    #------------------------------------------------------
    # Relationship / Reading
    #------------------------------------------------------

    def _fmt_Field(self, f: Field, level: int) -> str:
        t = self.format(f.type, level)
        return f"{f.name}:{t}"

    def _fmt_Relationship(self, r: Relationship, level: int) -> str:
        label = _relationship_label(r)
        if label:
            return label
        fields = ", ".join(self._fmt_Field(f, level) for f in r._fields)
        return f"rel({fields})"

    def _fmt_Reading(self, rd: Reading, level: int) -> str:
        label = _relationship_label(rd)
        if label:
            return label
        return rd._reading

    #------------------------------------------------------
    # Chains & Expressions
    #------------------------------------------------------

    def _fmt_Chain(self, p: Chain, level: int) -> str:
        start = self.format(p._start, level)
        nxt = _relationship_label(p._next, include_owner=False)
        if not nxt:
            nxt = self.format(p._next, level)
        return f"{start}.{nxt}"

    def _fmt_New(self, new: New, level: int) -> str:
        concept_label = self.format(new._op, level)
        op = f"{concept_label}.new"
        if new._identity_only:
            op += "_identity"
        arg_strs = [self.format(a, level) for a in new._args]
        kw_strs = [f"{k}={self.format(v, level)}" for k, v in (new._kwargs or {}).items()]
        parts = arg_strs + kw_strs

        if not parts:
            return op

        unit = self.indent_unit
        inline_pieces = [op] + parts
        inline_str = " ".join(inline_pieces)
        if (
            len(parts) <= self.inline_args_threshold
            and len(inline_str) + 2 <= self.max_inline_len
            and not any(_is_multiline(p) for p in parts)
        ):
            return f"({inline_str})"

        return _block(op, parts, level, unit, self.max_inline_len)

    def _fmt_Expression(self, e: Expression, level: int) -> str:
        op = self.format(e._op, level)
        # Prepare args/kwargs strings
        args = list(e._args)
        if isinstance(e._op, (Relationship, Reading)) and args:
            fields: list[Field] = getattr(e._op, "_fields", [])
            if fields:
                owner = fields[0].type
                if args and args[0] is owner:
                    args = args[1:]
        arg_strs = [self.format(a, level) for a in args]
        kw_strs = [f"{k}={self.format(v, level)}" for k, v in (e._kwargs or {}).items()]
        parts = arg_strs + kw_strs

        # No-arg expressions: treat as bare operator name
        if not parts:
            return op

        unit = self.indent_unit
        inline_pieces = [op] + parts
        inline_str = " ".join(inline_pieces)
        if (
            len(parts) <= self.inline_args_threshold
            and len(inline_str) + 2 <= self.max_inline_len
            and not any(_is_multiline(p) for p in parts)
        ):
            return f"({inline_str})"

        return _block(op, parts, level, unit, self.max_inline_len)

    #------------------------------------------------------
    # Match, Union, Not, Distinct
    #------------------------------------------------------

    def _fmt_Match(self, m: Match, level: int) -> str:
        unit = self.indent_unit
        items = [self.format(it, level + 1) for it in m._items]
        # inline if small & single-line
        inline = " | ".join(items)
        if len(inline) <= self.max_inline_len and all(not _is_multiline(s) for s in items):
            return f"(match {inline})"
        body = [s for s in items]
        return _block("match", body, level, unit)

    def _fmt_Union(self, u: Union, level: int) -> str:
        unit = self.indent_unit
        items = [self.format(it, level + 1) for it in u._items]
        inline = " & ".join(items)
        if len(inline) <= self.max_inline_len and all(not _is_multiline(s) for s in items):
            return f"(union {inline})"
        body = [s for s in items]
        return _block("union", body, level, unit)

    def _fmt_Not(self, n: Not, level: int) -> str:
        unit = self.indent_unit
        items = [self.format(it, level + 1) for it in n._items]
        inline = _join_inline(items)
        if len(inline) <= self.max_inline_len and all(not _is_multiline(s) for s in items):
            return f"(not {inline})"
        return _block("not", items, level, unit)

    def _fmt_Distinct(self, d: Distinct, level: int) -> str:
        unit = self.indent_unit
        items = [self.format(it, level + 1) for it in d._items]
        inline = _join_inline(items)
        if len(inline) <= self.max_inline_len and all(not _is_multiline(s) for s in items):
            return f"(distinct {inline})"
        return _block("distinct", items, level, unit)

    def _fmt_Group(self, group: Group, level: int) -> str:
        unit = self.indent_unit
        items = [self.format(it, level + 1) for it in group._args]
        if not items:
            return "(group)"
        inline = _join_inline(items)
        if len(inline) <= self.max_inline_len and all(not _is_multiline(s) for s in items):
            return f"(group {inline})"
        return _block("group", items, level, unit)

    def _fmt_Aggregate(self, agg: Aggregate, level: int) -> str:
        unit = self.indent_unit

        op_label = _relationship_label(agg._op, include_owner=False)
        if not op_label:
            op_label = self.format(agg._op, level)

        parts: list[str] = [self.format(arg, level + 1) for arg in agg._args]

        group_args = getattr(agg._group, "_args", [])
        if group_args:
            pretty_group = [self.format(x, level + 1) for x in group_args]
            inline = _join_inline(pretty_group)
            if len(inline) <= self.max_inline_len and all(not _is_multiline(s) for s in pretty_group):
                parts.append(f"(per {inline})")
            else:
                parts.append(_block("per", pretty_group, level, unit, self.max_inline_len))

        where_fragment = agg._where
        if where_fragment is not None:
            where_items = where_fragment._where
            if where_items:
                pretty_where = [self.format(x, level + 1) for x in where_items]
                inline = _join_inline(pretty_where)
                if len(inline) <= self.max_inline_len and all(not _is_multiline(s) for s in pretty_where):
                    parts.append(f"(where {inline})")
                else:
                    parts.append(_block("where", pretty_where, level, unit, self.max_inline_len))

        if not parts:
            return op_label

        inline_pieces = [op_label] + parts
        inline_str = " ".join(inline_pieces)
        if (
            len(parts) <= self.inline_args_threshold
            and len(inline_str) + 2 <= self.max_inline_len
            and not any(_is_multiline(p) for p in parts)
        ):
            return f"({inline_str})"

        return _block(op_label, parts, level, unit, self.max_inline_len)

    #------------------------------------------------------
    # Fragment
    #------------------------------------------------------

    def _fmt_Fragment(self, f: Fragment, level: int) -> str:
        """
        Renders (select ...)\n(where ...)\n(require ...)\n(then ...)\n(order_by ...)\n(limit N)
        Omit empty sections. Properly indents nested structures.
        """
        unit = self.indent_unit
        sections: list[str] = []

        def _fmt_list(name: str, items: list[Any]) -> None:
            if not items:
                return
            pretty_items = [self.format(x, level + 1) for x in items]
            # force multiline block semantics for Fragment sections
            sections.append(_block(name, pretty_items, level, unit, self.max_inline_len))

        _fmt_list("select", f._select)
        _fmt_list("require", f._require)
        _fmt_list("define",   f._define)
        _fmt_list("where", f._where)
        if f._order_by:
            # order_by: allow inline if all single-line and short
            pretty_items = [self.format(x, level + 1) for x in f._order_by]
            inline = _join_inline(pretty_items)
            if len(inline) <= self.max_inline_len and all(not _is_multiline(s) for s in pretty_items):
                sections.append(_indent(f"(order_by {inline})", level, unit))
            else:
                sections.append(_block("order_by", pretty_items, level, unit, self.max_inline_len))
        if getattr(f, "_limit", 0):
            sections.append(_indent(f"(limit {f._limit})", level, unit))

        sections = [s for s in sections if s]
        if not sections:
            return "(fragment)"
        return _block("fragment", sections, level, unit, self.max_inline_len)

    #------------------------------------------------------
    # Fallbacks for generic nodes
    #------------------------------------------------------

    def _fmt_Variable(self, v: Variable, level: int) -> str:
        # For unknown Variable subtypes, fall back to a safe identity-ish string
        # You can specialize more types above as needed.
        cls = v.__class__.__name__
        return f"<{cls} #{id(v):x}>"

#------------------------------------------------------
# Convenience API
#------------------------------------------------------

def format(obj: Any, **opts) -> str:
    """
    Pretty-format any DSL object.
    Example: print(format(fragment))
    """
    return PrettyPrinter(**opts).format(obj)

def pprint(obj: Any, **opts):
    """
    Pretty-print any DSL object.
    Example: print(pp(fragment))
    """
    print(format(obj, **opts))



import warnings

from v0.relationalai.semantics.rel.executor import RelExecutor

__all__ = ["RelExecutor"]

warnings.warn(
    "relationalai.early_access.rel.executor is deprecated, "
    "Please migrate to relationalai.semantics.rel.executor",
    DeprecationWarning,
    stacklevel=2,
)


from typing import Tuple
from v0.relationalai.semantics.lqp import ir as lqp

def mk_and(args: list[lqp.Formula]) -> lqp.Formula:
    # Flatten nested conjunctions
    if any(isinstance(arg, lqp.Conjunction) for arg in args):
        final_args = []
        for arg in args:
            if isinstance(arg, lqp.Conjunction):
                final_args.extend(arg.args)
            else:
                final_args.append(arg)
        args = final_args

    if len(args) == 1:
        return args[0]

    return lqp.Conjunction(args=args, meta=None)

def mk_or(args: list[lqp.Formula]) -> lqp.Formula:
    # Flatten nested disjunctions
    if any(isinstance(arg, lqp.Disjunction) for arg in args):
        final_args = []
        for arg in args:
            if isinstance(arg, lqp.Disjunction):
                final_args.extend(arg.args)
            else:
                final_args.append(arg)
        args = final_args

    if len(args) == 1:
        return args[0]

    return lqp.Disjunction(args=args, meta=None)

def mk_abstraction(vars: list[Tuple[lqp.Var, lqp.Type]], value: lqp.Formula) -> lqp.Abstraction:
    return lqp.Abstraction(vars=vars, value=value, meta=None)

def mk_exists(vars: list[Tuple[lqp.Var, lqp.Type]], value: lqp.Formula) -> lqp.Formula:
    if len(vars) == 0:
        return value
    abstr = mk_abstraction(vars, value)
    return lqp.Exists(body=abstr, meta=None)

def mk_specialized_value(value) -> lqp.SpecializedValue:
    return lqp.SpecializedValue(value=value, meta=None)

def mk_value(value) -> lqp.Value:
    return lqp.Value(value=value, meta=None)

def mk_type(typename: lqp.TypeName, parameters: list[lqp.Value]=[]) -> lqp.Type:
    return lqp.Type(type_name=typename, parameters=parameters, meta=None)

def mk_primitive(name: str, terms: list[lqp.RelTerm]) -> lqp.Primitive:
    return lqp.Primitive(name=name, terms=terms, meta=None)

def mk_pragma(name: str, terms: list[lqp.Var]) -> lqp.Pragma:
    return lqp.Pragma(name=name, terms=terms, meta=None)

def mk_attribute(name: str, args: list[lqp.Value]) -> lqp.Attribute:
    return lqp.Attribute(name=name, args=args, meta=None)

def mk_transaction(
    epochs: list[lqp.Epoch],
    configure: lqp.Configure = lqp.construct_configure({}, None),
    sync = None
) -> lqp.Transaction:
    return lqp.Transaction(epochs=epochs, configure=configure, sync=sync, meta=None)



from __future__ import annotations
from contextlib import contextmanager
from enum import Enum
from functools import lru_cache, wraps
from itertools import chain
from re import sub
from typing import Any, Optional, Sequence, cast
from collections.abc import Iterable

from ...util.naming import Namer, sanitize
from ...util.source import SourcePos
from ...util.structures import KeyedDict, KeyedSet, OrderedSet
from .pprint import pprint
from ..metamodel.builtins import builtins as bt
from ..metamodel.pprint import pprint as mmpp, print_tree
from ...util.error import Diagnostic, Part, err, warn, source as err_source, exc

from .base import (
    Alias, AsBool, CoreConcepts, CoreRelationships, DSLBase, Data, DerivedColumn, Distinct, Field, FieldRef, FilterBy, Group, Library, Literal, Model, Fragment, ModelEnum, New, Not, Chain,
    Property, Ref, Relationship, Concept, Expression, Statement, Table, TableSchema, TupleVariable, Value, Variable, Match, Union, dsl_key, is_primitive,
    NumberConcept, Reading, Aggregate, MetaRef
)
from ..metamodel.metamodel import (
    Aggregate as mAggregate, Annotation, Construct, Effect, FieldType, ISeq, ListType, Logical, Node, NoneType, NumberType, Relation, RelationType, Require, ScalarType, Overload,
    Lookup, TupleType, TypeNode, UnresolvedRelation, Var, Task, Literal as mLiteral, Value as mValue, Table as mTable, Update,
    Field as mField, Not as mNot, Reading as mReading, Match as mMatch, Union as mUnion, Data as mData, Model as mModel
)

#------------------------------------------------------
# Memoize
#------------------------------------------------------

def memoize_in_ctx(method):
    """Cache results in ctx.seen using id(node) as the key."""
    @wraps(method)
    def wrapper(self, ctx, node, *args, **kwargs):
        seen = ctx.seen
        key = id(node)
        if key in seen:
            return seen[key]
        seen[key] = result = method(self, ctx, node, *args, **kwargs)
        return result
    return wrapper

#------------------------------------------------------
# Helpers
#------------------------------------------------------

def find_keys(item: Value | Statement, root_only: bool = False) -> list[Value]:
    if isinstance(item, (Concept, Ref)):
        return [item]
    if isinstance(item, Chain):
        keys = []
        if not root_only and not isinstance(item._next, Property) and len(item._next._fields) > 1:
            keys.extend([FieldRef(item, ix + 1, field, ix + 1) for ix, field in enumerate(item._next._fields[1:-1])])
            keys.append(item)
        keys.extend(find_keys(item._start, root_only))
        return keys
    if isinstance(item, Relationship):
        if root_only:
            return [item]
        return list(item._fields[:-1])
    if isinstance(item, FieldRef):
        return find_keys(item._root, root_only)
    if isinstance(item, Expression):
        root = item._root if item._root is not None else item._op
        is_func = isinstance(root, Relationship) and any(field.is_input for field in root._fields)
        if root_only and item._root is not None and not(is_func):
            # We ignore "function-like relationships" (those with inputs) since they are never a root
            if isinstance(item._root, Relationship):
                return [item]
            return find_keys(item._root, root_only=True)
        args = item._args[:-1] if isinstance(item._op, Property) else item._args
        if is_func:
            assert isinstance(root, Relationship)
            args = [item._args[ix] for ix, field in enumerate(root._fields) if field.is_input]
        keys = []
        for arg in args:
            keys.extend(find_keys(arg, root_only))
        return keys
    if isinstance(item, Fragment):
        keys = []
        for subitem in item._select:
            keys.extend(find_keys(subitem, root_only))
        return keys
    if isinstance(item, DerivedColumn):
        return find_keys(item._table, root_only)
    if isinstance(item, Union):
        # Unions should always act as sets and so have no keys
        return []
    if isinstance(item, Match):
        keys = []
        if root_only:
            keys.extend(find_keys(item._items[0], root_only))
        else:
            for subitem in item._items:
                keys.extend(find_keys(subitem, root_only))
        return keys
    if isinstance(item, TupleVariable):
        keys = []
        for subitem in item._items:
            keys.extend(find_keys(subitem, root_only))
        return keys
    return []

def mm_value_type(value: mValue) -> TypeNode:
    if isinstance(value, (Var, mLiteral)):
        return value.type
    elif isinstance(value, TypeNode):
        return value
    elif isinstance(value, Relation):
        return RelationType
    elif isinstance(value, Field):
        return FieldType
    elif isinstance(value, Iterable):
        return TupleType(element_types=tuple(mm_value_type(v) for v in value))
    elif value is None:
        return NoneType
    else:
        raise NotImplementedError(f"mm_value_type not implemented for value type: {type(value)}")

@lru_cache(maxsize=1)
def _get_core():
    from . import core
    return core

@lru_cache(maxsize=1)
def _get_constraints():
    from ..std import constraints
    return constraints

def is_core_concept(concept):
    core = _get_core()
    if concept._model is core.Core:
        return True
    return any(is_core_concept(a) for a in concept._extends)

def concept_var_name(concept: Concept|Relationship) -> str:
    if isinstance(concept, Relationship):
        return sanitize(concept._short_name).lower()
    core = _get_core()
    if isinstance(concept, core.NumberConcept) or concept is core.Numeric:
        return "number"
    return sanitize(concept._name).lower()


#------------------------------------------------------
# Reference scheme handling
#------------------------------------------------------

def report_missing(concept:Concept, missing:OrderedSet[str], required:OrderedSet[str], source, ctx:Context):
    if missing:
        missing_str = ", ".join([f"`{name}`" for name in missing])
        required_str = ", ".join([f"[bold red]{name}[/bold red]" if name in missing else name for name in required])
        message = "values for properties" if len(missing) > 1 else "a value for property"
        ctx.err("Missing identity", f"Missing {message} {missing_str} in concept `{concept._name}`.", [
            err_source(source),
            f"{concept._name} requires {required_str}"
        ])
        return True
    return False

def make_construct(compiler:FrontCompiler,ctx:Context, concept:Concept, kwargs:dict[str, Any], ident_args:list[str], target:Var, source:Any=None):
    construct_args:list[mValue] = [
        ctx.to_value(concept._name),
        *[compiler.lookup(ctx, kwargs[ident]) for ident in ident_args]
    ]
    return Construct(tuple(construct_args), target, source=source)

def handle_concept_expression(compiler:FrontCompiler, ctx:Context, expr:Expression) -> Var:
    assert isinstance(expr._op, Concept)
    source = expr._source
    passed_concept = expr._op
    kwargs = {k.lower(): v for k, v in expr._kwargs.items()}

    root_concept = None
    hierarchy:list[Concept] = []
    constructs = KeyedDict[Concept, list[str]](dsl_key)
    required = OrderedSet[str]()
    missing = OrderedSet[str]()

    def walk_extends(c:Concept):
        nonlocal root_concept
        if c in hierarchy:
            return
        for ext in c._extends:
            walk_extends(ext)
        hierarchy.append(c)
        if c._identify_by and root_concept is None:
            root_concept = c
        ident_args = []
        for ident_rel in c._identify_by:
            ident = ident_rel._short_name.lower()
            required.add(ident)
            if ident in kwargs:
                ident_args.append(ident)
            else:
                missing.add(ident)
        constructs[c] = ident_args

    walk_extends(passed_concept)

    if root_concept is None:
        root_concept = hierarchy[0]
        if isinstance(expr, New) and expr._row_ids:
            names = [f"_row_id_{t._name}" for t in expr._row_ids]
            kwargs.update({name: expr._row_ids[ix] for ix, name in enumerate(names)})
            constructs[root_concept] = list(sorted(names))
        else:
            constructs[root_concept] = list(sorted(kwargs.keys()))

    root_id = ctx.to_value(expr)
    assert isinstance(root_id, Var)
    root_args = constructs[root_concept]

    if isinstance(expr, New) and ctx.in_update:
        # New requires that none of the identity args are missing
        if report_missing(passed_concept, missing, required, source, ctx):
            return root_id

        # construct the root identity
        ctx.add(make_construct(compiler, ctx, root_concept, kwargs, root_args, root_id, source=source))
        if expr not in ctx.construct_handled:
            ctx.add_update(passed_concept, [root_id], source=source)

            # Update identity and non-identity attributes
            for k, v in kwargs.items():
                with ctx.subcontext():
                    rel = passed_concept._dot(k)
                    ctx.add_update(rel, [root_id, compiler.lookup(ctx, v)], source=source)
                    nodes = tuple(ctx.nodes)
                ctx.add(Logical(nodes, source=source, optional=True))

            ctx.construct_handled.add(expr)

    elif isinstance(expr, New):
        # New requires that none of the identity args are missing
        if report_missing(passed_concept, missing, required, source, ctx):
            return root_id

        # just construct the root identity and then look it up in the subtype
        # population
        ctx.add(make_construct(compiler, ctx, root_concept, kwargs, root_args, root_id, source=source))
        ctx.add_lookup(passed_concept, [root_id], source=source)

        # Any non-identity attributes are filters
        for k, v in kwargs.items():
            if k not in root_args:
                rel = passed_concept._dot(k)
                ctx.add_lookup(rel, [root_id, compiler.lookup(ctx, v)], source=source)

    elif isinstance(expr, FilterBy):
        # check if _any_ concept has all the identity args it needs as a
        if len(root_args) == len(root_concept._identify_by):
            ctx.add(make_construct(compiler, ctx, root_concept, kwargs, root_args, root_id, source=source))
        else:
            # if we don't have all the identity args for the root, we need to look up any partials
            root_args = []

        ctx.add_lookup(passed_concept, [root_id], source=source)
        for k, v in kwargs.items():
            if k not in root_args:
                rel = passed_concept._dot(k)
                ctx.add_lookup(rel, [root_id, compiler.lookup(ctx, v)], source=source)
    else:
        exc("Invalid concept expression", f"Expected a `New` or `FilterBy` expression, got `{type(expr).__name__}`.", [err_source(source)])

    return root_id

#------------------------------------------------------
# Match union branch keys
#------------------------------------------------------

def get_branch_keys(compiler:FrontCompiler, ctx:Context, branch:Match|Union) -> list[tuple[Value, Var]]:
    keys = KeyedSet(dsl_key)
    for item in branch._items:
        if not isinstance(item, Fragment):
            keys.update(find_keys(item))
    branch_keys = []
    seen_keys = set()
    for key in keys:
        k_var = ctx.to_value(key)
        if k_var not in seen_keys and isinstance(k_var, Var) and not ctx.is_var_available(k_var, ignore_current=True):
            seen_keys.add(k_var)
            branch_keys.append((key, k_var))
    return branch_keys

def normalize_branch_keys(compiler:FrontCompiler, ctx:Context, branch_keys:list[tuple[Value, Var]], branch:Statement):
    for key, var in branch_keys:
        if var not in ctx.frame_vars:
            source = branch._source if isinstance(branch, DSLBase) else var.source
            if not isinstance(key, (Concept, Ref)):
                return ctx.err("Invalid branch", "Branch key must be a concept or reference.", [err_source(source)])
            concept = key._to_concept()
            if is_core_concept(concept):
                return ctx.err("Invalid branch", "Branch key must be an entity, not a primitive.", [err_source(source)])
            else:
                ctx.add(Construct((ctx.to_value(concept._name + "_NONE_SENTINEL"),), var, source=key._source))


#------------------------------------------------------
# Context
#------------------------------------------------------

class Frame:
    def __init__(self):
        self.nodes:OrderedSet[Task] = OrderedSet()
        self.in_update = False
        self.used_vars:OrderedSet[Var] = OrderedSet()
        self.seen:dict[Value, mValue] = {}

class Context:
    def __init__(self, compiler:FrontCompiler):
        self.compiler = compiler
        self.stack:list[Frame] = [Frame()]
        self.value_map:KeyedDict[Value|tuple[Value, Value], mValue] = KeyedDict(dsl_key)
        self.has_error = False
        self._in_update = False
        self.construct_handled:KeyedSet[New] = KeyedSet(dsl_key)

    #------------------------------------------------------
    # subcontext
    #------------------------------------------------------

    @contextmanager
    def subcontext(self):
        self.stack.append(Frame())
        try:
            yield
        finally:
            prev = self.stack.pop()
            self.stack[-1].used_vars.update(prev.used_vars)

    #------------------------------------------------------
    # in_update
    #------------------------------------------------------

    @contextmanager
    def updating(self):
        prev = self.in_update
        self._in_update = True
        try:
            yield
        finally:
            self._in_update = prev
            if not self._in_update:
                self.construct_handled.clear()

    @property
    def in_update(self) -> bool:
        return self._in_update

    #------------------------------------------------------
    # Errors/warnings
    #------------------------------------------------------

    def err(self, name: str, message: str, parts: Optional[list[Part]] = None) -> Diagnostic:
        self.has_error = True
        return err(name, message, parts)

    def warn(self, name: str, message: str, parts: Optional[list[Part]] = None) -> Diagnostic:
        self.has_error = True
        return warn(name, message, parts)

    #------------------------------------------------------
    # Task capture
    #------------------------------------------------------

    def _check_frames(self, node:Task):
        for frame in reversed(self.stack[:-1]):
            if node in frame.nodes:
                return True
        return False

    def add(self, node:Task):
        if self._check_frames(node):
            return
        self.stack[-1].nodes.add(node)

    def extend(self, nodes:Sequence[Task]):
        for node in nodes:
            self.add(node)

    def add_lookup(self, relationship:Relationship|Concept, args:Sequence[mValue], source:Any=None):
        op = self.compiler.to_relation(relationship)
        hint = None
        if isinstance(relationship, Reading):
            hint = self.compiler.find_reading(relationship)
            # reorder the args to match the correct ordering
            new_args:list[Any] = [None] * len(hint.field_order)
            for cur, new in enumerate(hint.field_order):
                new_args[new] = args[cur]
            args = new_args
        self.add(Lookup(op, tuple(args), reading_hint=hint, source=source))

    def add_update(self, relationship:Relationship|Concept, args:Sequence[mValue], effect=Effect.derive, source:Any=None):
        op = self.compiler.to_relation(relationship)
        hint = None
        if isinstance(relationship, Reading):
            hint = self.compiler.reading_map[relationship]
            # reorder the args to match the correct ordering
            new_args:list[Any] = [None] * len(hint.field_order)
            for cur, new in enumerate(hint.field_order):
                new_args[new] = args[cur]
            args = new_args
        self.add(Update(op, tuple(args), effect=effect, reading_hint=hint, source=source))

    def add_eq(self, left:mValue, right:mValue, source:Any=None):
        self.add(Lookup(getattr(bt.core, "="), (left, right), source=source))

    @property
    def nodes(self):
        return self.stack[-1].nodes

    #------------------------------------------------------
    # Value mapping
    #------------------------------------------------------

    def to_value(self, value:Value, source:Optional[Value]=None) -> mValue:
        key = (source, value) if source is not None else value
        if res := self.value_map.get(key):
            if isinstance(res, Var):
                self.stack[-1].used_vars.add(res)
            return res

        node = None
        if isinstance(value, Concept):
            value_type = self.compiler.to_type(value)
            node = Var(value_type, concept_var_name(value), source=value._source)
        elif isinstance(value, Chain):
            node = self.to_value(value._next._fields[-1], value)
        elif isinstance(value, Ref):
            value_type = self.compiler.to_type(value._concept)
            name = sanitize(value._name).lower() if value._name else f"{concept_var_name(value._concept)}{value._id}"
            node = Var(value_type, name, source=value._source)
        elif isinstance(value, Field):
            value_type = self.compiler.to_type(value.type)
            node = Var(value_type, sanitize(value.name).lower())
        elif isinstance(value, New):
            value_type = self.compiler.to_type(value._op)
            node = Var(value_type, f"{concept_var_name(value._op)}", source=value._source)
        elif isinstance(value, FieldRef):
            node = self.to_value(value._resolved, source=value._root)
        elif isinstance(value, Expression):
            node = self.to_value(value._args[-1])
        elif isinstance(value, Aggregate):
            node = self.to_value(value._args[-1])
        elif isinstance(value, DerivedColumn):
            value_type = self.compiler.to_type(value._type) if value._type is not None else bt.core.Any
            node = Var(value_type, f"v", source=value._source)
        elif isinstance(value, Literal):
            value_type = self.compiler.to_type(value._type)
            node = mLiteral(type=value_type, value=value._value, source=value._source)
        elif isinstance(value, AsBool):
            node = Var(bt.core.Boolean, f"v", source=value._source)
        elif isinstance(value, MetaRef):
            if isinstance(value._target, Concept):
                node = self.compiler.to_type(value._target)
            elif isinstance(value._target, Relationship):
                node = self.compiler.to_relation(value._target)
            elif isinstance(value._target, Field):
                node = self.compiler.to_field(value._target)
            else:
                node = self.to_value(value._target)
        elif is_primitive(value):
            type_ = self.compiler.to_type(Literal._get_type(value))
            node = mLiteral(type=type_, value=value)
        else:
            raise NotImplementedError(f"to_value not implemented for value type: {type(value)}")

        if isinstance(node, Var):
            self.stack[-1].used_vars.add(node)
        self.value_map[key] = node
        return node

    @property
    def frame_vars(self) -> OrderedSet[Var]:
        return self.stack[-1].used_vars

    def is_var_available(self, var:Var, ignore_current=False) -> bool:
        stack = self.stack[:-1] if ignore_current else self.stack
        for frame in reversed(stack):
            if var in frame.used_vars:
                return True
        return False

    @property
    def seen(self) -> dict[Value, mValue]:
        return self.stack[-1].seen

#------------------------------------------------------
# Compiler
#------------------------------------------------------

class FrontCompiler:
    def __init__(self, model: Model):
        self.model = model
        self.relations:KeyedDict[DSLBase, Relation] = KeyedDict(dsl_key)
        self.fields:KeyedDict[Field, mField] = KeyedDict(dsl_key)
        self.reading_map: KeyedDict[Reading, mReading] = KeyedDict(dsl_key)
        self.imported_libraries: set[Model] = set()
        self.relation_constraints: OrderedSet[Lookup] = OrderedSet()
        self.global_namer = Namer()

    #------------------------------------------------------
    # Library handling
    #------------------------------------------------------

    def check_import(self, node: Any):
        if not isinstance(node, DSLBase):
            return
        if isinstance(node, Aggregate):
            self.check_import(node._op)
        model = node._model
        if model and model != self.model and model not in self.imported_libraries:
            if model not in self.model.libraries:
                self.model.libraries.append(model)
            self.imported_libraries.add(model)
            self.relations.update(model._compiler.relations)
            self.relation_constraints.update(model._compiler.relation_constraints)
            return True

    #------------------------------------------------------
    # Model elements
    #------------------------------------------------------

    def find_reading(self, reading:Reading) -> mReading:
        if found := self.reading_map.get(reading):
            return found
        # It's possible that readings were added after the first time we encounter
        # the relationship, so we need to re-create them here and update the relationship
        root = reading._relationship
        rel = self.to_relation(root)
        readings = self.to_readings(root, start=len(rel.readings))
        rel._dangerous_set_readings(rel.readings + readings)
        return self.reading_map[reading]

    def to_readings(self, relationship: Relationship, start=0) -> tuple[mReading, ...]:
        readings = tuple(mReading(name=reading._short_name, parts=tuple(reading._parts), source=reading._source) for reading in relationship._readings[start:])
        self.reading_map.update((reading, mreading) for reading, mreading in zip(relationship._readings[start:], readings))
        return readings

    def to_annotation(self, node: Expression | Relationship) -> Annotation:
        if isinstance(node, Expression):
            args = []
            for arg in node._args:
                if isinstance(arg, Literal):
                    args.append(mLiteral(type=self.to_type(arg._type), value=arg._value, source=arg._source))
                elif isinstance(arg, Concept) or isinstance(arg, Relationship):
                    args.append(self.to_relation(arg))
                elif isinstance(arg, Chain):
                    args.append(self.to_relation(arg._next))
                else:
                    exc("Invalid annotation", f"Invalid annotation argument type: {type(arg).__name__}", [
                        err_source(arg._source),
                        "Only Concepts, Relationships, and Literals are allowed as annotation arguments."
                    ])
            return Annotation(relation=self.to_relation(node._op), args=tuple(args), source=node._source)
        elif isinstance(node, Relationship):
            return Annotation(relation=self.to_relation(node), args=(), source=node._source)
        else:
            raise ValueError(f"Unknown annotation type: {type(node)} - {node}")

    def to_field(self, node: Field) -> mField:
        if res := self.fields.get(node):
            return res
        field_type = self.to_type(node.type)
        if node.is_list:
            field_type = ListType(element_type=field_type)
        field = mField(name=node.name, type=field_type, input=node.is_input, source=node._source)
        self.fields[node] = field
        return field

    def to_relation(self, node: Any) -> Relation:
        if res := self.relations.get(node):
            return res
        elif self.check_import(node) and (res := self.relations.get(node)):
            return res

        rel = None
        if isinstance(node, Reading):
            rel = self.to_relation(node._relationship)
        elif isinstance(node, Relationship):
            annos = tuple(self.to_annotation(anno) for anno in node._annotations)
            readings = self.to_readings(node)
            fields = [self.to_field(field) for field in node._fields]
            if node._short_name == '':
                name = self.global_namer.get_name(sub(r"[{}: ]", "_", str(node._readings[0]._reading)).strip("_"))
            else:
                name = node._short_name
            overloads = []
            if node._overloads:
                for overload in node._overloads:
                    overload_types = tuple(self.to_type(concept) for concept in overload)
                    overloads.append(Overload(overload_types))
            if node._is_unresolved:
                rel = UnresolvedRelation(name=name, fields=tuple(fields), source=node._source, readings=readings, annotations=annos, overloads=tuple(overloads))
            else:
                rel = Relation(name=name, fields=tuple(fields), source=node._source, readings=readings, annotations=annos, overloads=tuple(overloads))
            if isinstance(node, Property):
                # Add a uniqueness constraint for properties
                _get_constraints()
                self.relation_constraints.add(Lookup(bt.constraints.unique_fields, (tuple(fields[:-1]),)))
        elif isinstance(node, Data):
            rel = mData(name=f"Data{node._id}", data=node._data, source=node._source)
            self.relations[node] = rel
            rel._force_columns(tuple(self.to_relation(col) for col in node._columns))
        elif isinstance(node, Table):
            rel = mTable(name=node._name, source=node._source)
            self.relations[node] = rel
            rel._force_columns(tuple(self.to_relation(col) for col in node._columns))
        elif isinstance(node, NumberConcept):
            super_types=tuple(self.to_type(ancestor) for ancestor in node._extends)
            rel = NumberType(name=node._name, precision=node._precision, scale=node._scale, source=node._source, super_types=super_types)
        elif isinstance(node, Concept):
            annos = tuple(self.to_annotation(anno) for anno in node._annotations)
            super_types=tuple(self.to_type(ancestor) for ancestor in node._extends)
            rel = ScalarType(name=node._name, annotations=annos, source=node._source, super_types=super_types)
            self.relations[node] = rel
            identify_by_rels = tuple(self.to_relation(rel) for rel in node._identify_by)
            rel._force_identify_by(identify_by_rels)
        else:
            raise ValueError(f"Unknown node type: {type(node)} - {node}")
        self.relations[node] = rel
        return rel

    def to_type(self, node: Any) -> ScalarType:
        rel = self.to_relation(node)
        if not isinstance(rel, ScalarType):
            raise ValueError(f"Node is not a ScalarType: {node}")
        return rel

    #------------------------------------------------------
    # Compile
    #------------------------------------------------------

    def compile(self, fragment: Fragment) -> Task|list[Task]:
        ctx = Context(self)
        # pprint(fragment)
        return self.fragment(ctx, fragment)

    def compile_model(self, model: Model) -> mModel:
        def as_seq(x):
            return x if isinstance(x, (list, tuple)) else (x,)

        # init enums
        for enum in model.enums:
            enum._init_members()

        self.check_invariants(model)

        for concept in model.concepts:
            self.to_relation(concept)
        for relationship in model.relationships:
            self.to_relation(relationship)
        for table in model.tables:
            self.to_relation(table)

        # compile + flatten defines/requires in one pass
        compiled:list[Task] = []
        compiled.extend(chain.from_iterable(
            as_seq(self.compile(item)) for item in chain(model.defines, model.requires)
        ))
        # partition relation values into types vs non-types
        rel_vals = list(self.relations.values())
        types = tuple(r for r in rel_vals if isinstance(r, TypeNode))
        relations = tuple(r for r in rel_vals if not isinstance(r, TypeNode))

        compiled.insert(0, Require(check=Logical(tuple(self.relation_constraints), source=model._source)))

        source = model._source
        root = Logical(tuple(compiled), source=source)

        return mModel(
            reasoners=(),
            relations=relations,
            types=types,
            root=root,
            source=source,
        )

    #------------------------------------------------------
    # Invariants
    #------------------------------------------------------

    def check_invariants(self, model: Model):
        # Ensure that all concepts with identify_by have at least one relationship
        has_errors = False
        for concept in model.concepts:
            for k, v in concept._relationships.items():
                for ext in concept._extends:
                    if k in ext._relationships:
                        kind = "Property" if isinstance(v, Property) else "Relationship"
                        parent_kind = "property" if isinstance(ext._relationships[k], Property) else "relationship"
                        err(f"{kind} conflict",
                            f"{kind} `{k}` in concept `{concept._name}` conflicts with {parent_kind} in parent concept `{ext._name}`.",
                            [
                                f"Parent {parent_kind}",
                                err_source(ext._relationships[k]),
                                f"Child {kind.lower()}",
                                err_source(v)
                            ])
                        has_errors = True

        if has_errors:
            exc("Invalid model", "Model has invariant violations, see errors for details.")

    #------------------------------------------------------
    # Fragment
    #------------------------------------------------------

    def fragment(self, ctx: Context, fragment: Fragment):
        if fragment._require:
            return self.require(ctx, fragment, scope=True)
        else:
            annos = tuple(self.to_annotation(anno) for anno in fragment._annotations)
            self.where(ctx, fragment, fragment._where)
            self.select(ctx, fragment, fragment._select)
            self.define(ctx, fragment, fragment._define)
            return Logical(tuple(ctx.nodes), scope=True, annotations=annos, source=fragment._source)

    #------------------------------------------------------
    # Where
    #------------------------------------------------------

    def where(self, ctx: Context, fragment:Fragment, exprs: list[Value]):
        for expr in exprs:
            self.lookup(ctx, expr)

    #------------------------------------------------------
    # Select
    #------------------------------------------------------

    def select(self, ctx: Context, fragment:Fragment, exprs: Sequence[Value]):
        if not exprs:
            return

        source = fragment._source
        output_name = f"output{fragment._id}"

        uri = f"dataframe://{output_name}" if fragment._into is None else f"table://{fragment._into._name}"
        table = mTable(name=f"Output{fragment._id}", uri=uri, source=source)
        table_var = Var(type=table, name=output_name, source=source)

        keys:KeyedSet[Value] = KeyedSet(dsl_key)
        items:list[Task] = []

        # Unwrap distinct if present
        is_distinct = any(isinstance(expr, Distinct) for expr in exprs)
        if is_distinct:
            if len(exprs) != 1 or not isinstance(exprs[0], Distinct):
                return ctx.err("Invalid distinct", "Distinct must be applied to the entire select.", [
                    err_source(source)
                ])
            exprs = exprs[0]._items

        # Build out the table columns
        # v0, v1, etc, for generic elements like literals or the unnamed result of union/matches
        range_namer = Namer(range=True)
        # v, v_2, v_3, etc, for elements where we want to preserve the original name if there
        # are no collisions, like chain expressions, aggregates, etc
        namer = Namer()

        # we have to first find all the roots, the roots need to be looked up in the outer scope
        # we then need to find the keys for each expr, any keys that aren't roots need to be added
        # to the column relation
        roots = KeyedSet(dsl_key)
        for expr in exprs:
            if is_distinct:
                roots.add(expr)
            else:
                roots.update(find_keys(expr, root_only=True))

        # look the roots up in the outer scope
        root_vars:OrderedSet[mValue] = OrderedSet()
        for root in roots:
            root_v = self.lookup(ctx, root)
            # For top-level relationships, the fields of the relationship are
            # the root vars, since all of them behave like keys. For Properties,
            # we ignore the last field since it's functionally determined by the others.
            # Expressions use their args
            if isinstance(root, Property):
                root_vars.update(ctx.to_value(field) for field in root._fields[:-1])
            elif isinstance(root, Relationship):
                root_vars.update(ctx.to_value(field) for field in root._fields)
            elif isinstance(root, Expression):
                root_vars.update(ctx.to_value(arg) for arg in root._args)
            else:
                root_vars.add(root_v)

        # handle order_by/limit
        if fragment._order_by or fragment._limit != 0:
            from ..std.aggregates import _sort_agg
            args = fragment._order_by if fragment._order_by else fragment._select
            sort_agg = _sort_agg(fragment._limit, *args)
            self.lookup(ctx, sort_agg)

        # build columns
        table_cols = []
        for expr in exprs:
            expr_keys = [expr] if is_distinct else find_keys(expr)

            with ctx.subcontext():
                col = self.lookup(ctx, expr)
                # Columns may need their own keys beyond the roots to guarantee that multiple values end up
                # available for the same root. E.g. people having multiple pets
                maybe_col_keys = [ctx.to_value(col_key) for col_key in expr_keys]
                col_keys = [col_key for col_key in maybe_col_keys
                            if isinstance(col_key, Var) and col_key != col and col_key not in root_vars]
                col_key_fields = [mField(name=key.name, type=key.type) for key in col_keys]

                # build the relation
                col_name = (expr._alias if isinstance(expr, Alias) else
                            range_namer.get_name(col.name) if isinstance(col, Var) and isinstance(expr, (Union, Match)) else
                            namer.get_name(self.to_relation(expr._op).name) if isinstance(expr, Aggregate) else
                            namer.get_name(col.name) if isinstance(col, Var) else
                            range_namer.get_name("v"))
                col_type = mm_value_type(col)
                col_relation = Relation(col_name, (
                    mField(name="table", type=table),
                    *col_key_fields,
                    mField(name=col_name, type=col_type)
                ), source=source)
                table_cols.append(col_relation)

                # The shim needs a hint that this output is both a key and a value
                annos = ()
                if col in maybe_col_keys and col not in root_vars:
                    annos = (Annotation(relation=bt.core["output_value_is_key"], args=(), source=source),)
                ctx.add(Update(col_relation, args=tuple([table_var, *col_keys, col]), annotations=annos, source=source))

                optional = True
                # rank and limit should always just get pushed to the root - they can never be null
                if isinstance(expr, Aggregate):
                    op = self.to_relation(expr._op)
                    optional = not (op is bt.aggregates.rank or op is bt.aggregates.limit)

                node_tuple = tuple(ctx.nodes)
                if len(ctx.nodes) == 1:
                    items.append(node_tuple[0])
                else:
                    items.append(Logical(node_tuple, source=source, optional=optional))

        # add the columns to the output table
        table._force_columns(tuple(table_cols))

        # Construct followed by adding the row to the table and then each column
        ctx.add(Construct(tuple(root_vars), table_var, source=source))
        ctx.add(Update(table, tuple([table_var]), source=source))
        ctx.extend(items)

    #------------------------------------------------------
    # Define
    #------------------------------------------------------

    def define(self, ctx: Context, fragment:Fragment, exprs: list[Value]):
        with ctx.updating():
            for expr in exprs:
                source = expr._source if isinstance(expr, DSLBase) else fragment._source
                with ctx.subcontext():
                    self.update(ctx, expr, root_source=fragment._source)
                    nodes = tuple(ctx.nodes)
                ctx.add(Logical(nodes, source=source, optional=True))

    #------------------------------------------------------
    # Require
    #------------------------------------------------------

    def require(self, ctx: Context, fragment:Fragment, scope:bool=False):
        source = fragment._source
        reqs = []
        annos = tuple(self.to_annotation(anno) for anno in fragment._annotations)
        with ctx.subcontext():
            self.where(ctx, fragment, fragment._where)
            domain_body = tuple(ctx.nodes)
            domain_vars = ctx.frame_vars
            for expr in fragment._require:
                with ctx.subcontext():
                    self.lookup(ctx, expr)
                    check = Logical(tuple(ctx.nodes), source=source)
                    hoisted = tuple(v for v in ctx.frame_vars if v in domain_vars)
                    reqs.append(Require(Logical(domain_body, source=source), check, scope=scope, annotations=annos, source=source))
        if len(reqs) == 1:
            return reqs[0]
        return reqs

    #------------------------------------------------------
    # Lookup
    #------------------------------------------------------

    # @memoize_in_ctx
    def lookup(self, ctx: Context, node: Any) -> mValue:
        source = node._source if not is_primitive(node) else None
        self.check_import(node)

        #------------------------------------------------------
        # Concept / Concept expressions (includes Data and Table)
        #------------------------------------------------------

        if isinstance(node, Concept):
            final_var = ctx.to_value(node)
            if not is_core_concept(node):
                ctx.add_lookup(node, [final_var], source=source)
            return final_var

        elif isinstance(node, Ref):
            final_var = ctx.to_value(node)
            if not is_core_concept(node._concept):
                ctx.add_lookup(node._concept, [final_var], source=source)
            return final_var

        elif isinstance(node, New):
            # If we encounter a New nested inside of an update, we want to treat this
            # new as an update as well so that you can write Person.new(.., pet=Pet.new(..))
            if ctx.in_update and not node._identity_only:
                return self.update(ctx, node, root_source=node._source)

            # Otherwise we're just checking this exists
            return handle_concept_expression(self, ctx, node)

        elif isinstance(node, FilterBy):
            return handle_concept_expression(self, ctx, node)

        #------------------------------------------------------
        # Relationship
        #------------------------------------------------------

        elif isinstance(node, Relationship):
            # we need vars for each field, but how do we store them?
            args = [ctx.to_value(field) for field in node._fields]
            ctx.add_lookup(node, args, source=source)
            return args[-1]

        elif isinstance(node, FieldRef):
            self.lookup(ctx, node._root)
            # If the source of the field ref is an expression, this is an ArgumentRef, so
            # return the expression var directly
            if isinstance(node._root, Expression):
                return self.lookup(ctx, node._root._args[node._resolved_ix])
            # If this is a reference to the relationship itself, then we want that to
            # unify with any other references to the given field. If it's through a chain
            # or expression, then it should be sourced from there.
            field_source = node._root if not isinstance(node._root, Relationship) else None
            return ctx.to_value(node._resolved, source=field_source)

        #------------------------------------------------------
        # Expressions
        #------------------------------------------------------

        elif isinstance(node, Expression):
            # For auto-filled expressions, there's no reason to lookup the last ref and
            # do a population check
            if node._auto_filled:
                args = [self.lookup(ctx, arg) for arg in node._args[:-1]] + [ctx.to_value(node._args[-1])]
                ctx.add_lookup(node._op, args, source=source)
                return args[-1]
            # if we are casting e.g. with something like `Discount(0.6)` we should return a literal with that
            # type
            elif isinstance(node._op, Concept) and is_core_concept(node._op):
                if isinstance(node._args[0], Literal):
                    return mLiteral(type=self.to_type(node._op), value=node._args[0]._value, source=source)
                else:
                    cur = self.lookup(ctx, node._args[-1])
                    ret = ctx.to_value(node._op, source=node)
                    ctx.add_lookup(CoreRelationships["cast"], [self.to_type(node._op), cur, ret], source=source)
                    return ret
            # Normal expression
            else:
                args = [self.lookup(ctx, a) for a in node._args]
                ctx.add_lookup(node._op, args, source=source)
                return args[-1]

        elif isinstance(node, Chain):
            start = self.lookup(ctx, node._start)
            final_var = ctx.to_value(node)
            # Unaries are special, the chain's var is just the start var
            if len(node._next._fields) == 1:
                ctx.add_lookup(node._next, [start], source=source)
                return start
            else:
                # If the relation has more than 2 fields, we need to create vars for the
                # unreferrenced ones in the middle
                middle_args = [ctx.to_value(field, source=node) for field in node._next._fields[1:-1]]
                args = [start, *middle_args, final_var]
                ctx.add_lookup(node._next, args, source=source)
            return final_var

        #------------------------------------------------------
        # Aggregates
        #------------------------------------------------------

        elif isinstance(node, Aggregate):
            agg = self.to_relation(node._op)
            input_args = [arg for ix, arg in enumerate(node._args) if agg.fields[ix].input]

            proj_set:KeyedSet[Value] = KeyedSet(dsl_key)
            if not node._distinct:
                for item in [*node._projection_args, *input_args]:
                    proj_set.update(find_keys(item))
            proj_set.update(node._projection_args)
            proj = cast(OrderedSet[Var], OrderedSet([ctx.to_value(p) for p in proj_set]))
            group = cast(list[Var], [self.lookup(ctx, arg) for arg in node._group._args])

            if agg is bt.aggregates.rank:
                assert isinstance(node._args[0], TupleVariable)
                for arg in node._args[0]._items:
                    self.lookup(ctx, arg)
            elif agg is bt.aggregates.limit:
                assert isinstance(node._args[1], TupleVariable)
                for arg in node._args[1]._items:
                    self.lookup(ctx, arg)

            with ctx.subcontext():
                for item in node._where._where:
                    self.lookup(ctx, item)
                for projected in proj_set:
                    self.lookup(ctx, projected)
                args = cast(list[Var], [self.lookup(ctx, arg) for arg in node._args])
                where = Logical(tuple(ctx.nodes), source=source)

            ctx.add(mAggregate(agg, tuple(proj), tuple(group), tuple(args), where, source=source))
            return args[-1] if not isinstance(args[-1], tuple) else None

        elif isinstance(node, Group):
            for arg in node._args:
                self.lookup(ctx, arg)
            return

        #------------------------------------------------------
        # Quantifiers
        #------------------------------------------------------

        elif isinstance(node, Not):
            with ctx.subcontext():
                for item in node._items:
                    self.lookup(ctx, item)
                not_ = mNot(Logical(tuple(ctx.nodes)), source=source)
            ctx.add(not_)
            return

        #------------------------------------------------------
        # Nested fragment
        #------------------------------------------------------

        elif isinstance(node, Fragment):
            if node._require or node._define:
                ctx.err("Invalid subquery", "Nested queries with `require` or `define` are not supported.", [
                    err_source(source)
                ])
            if node._select:
                with ctx.subcontext():
                    for where in node._where:
                        self.lookup(ctx, where)
                    # cols = cast(ISeq[Var], tuple(ctx.to_value(col) for col in node._columns))
                    vals = [self.lookup(ctx, sel) for sel in node._select]
                    ctx.value_map.update(zip(node._columns, vals))
                    # for col, val in zip(cols, vals):
                    #     ctx.add_eq(col, val, source=source)
                    nodes = tuple(ctx.nodes)
                ctx.add(Logical(nodes, source=source))
                return vals[-1]
            else:
                for where in node._where:
                    self.lookup(ctx, where)
                return None

        #------------------------------------------------------
        # Match / Union
        #------------------------------------------------------

        elif isinstance(node, (Match, Union)):
            branches = []
            outputs = cast(ISeq[Var], tuple(ctx.to_value(col) for col in node._columns) if node._arg_count() > 0 else tuple())
            branch_keys = get_branch_keys(self, ctx, node)
            for item in node._items:
                item_source = item._source if isinstance(item, Variable) else source
                if isinstance(item, Fragment):
                    if item._select:
                        # map the select columns to the match/union outputs
                        for ix, col in enumerate(item._columns):
                            ctx.value_map[col] = outputs[ix]
                        with ctx.subcontext():
                            self.lookup(ctx, item)
                            # unwrap the nested logical
                            wrapper = ctx.nodes[0]
                            assert isinstance(wrapper, Logical)
                            for subnode in wrapper.body:
                                ctx.add(subnode)

                            normalize_branch_keys(self, ctx, branch_keys, item)
                            # add the output column equalities
                            for ix, col in enumerate(item._columns):
                                ctx.add_eq(outputs[ix], ctx.to_value(col), source=item_source)
                            branches.append(Logical(tuple(ctx.nodes)[1:], source=item_source))
                    else:
                        with ctx.subcontext():
                            for where in item._where:
                                self.lookup(ctx, where)
                            branches.append(Logical(tuple(ctx.nodes), source=item_source))
                else:
                    with ctx.subcontext():
                        res = self.lookup(ctx, item)
                        normalize_branch_keys(self, ctx, branch_keys, item)
                        if outputs:
                            ctx.add_eq(outputs[0], res, source=item_source)
                        branches.append(Logical(tuple(ctx.nodes), source=item_source))

            if isinstance(node, Union):
                ctx.add(mUnion(tuple(branches), source=source))
            else:
                ctx.add(mMatch(tuple(branches), source=source))

            return outputs[-1] if outputs else None

        #------------------------------------------------------
        # Value-likes
        #------------------------------------------------------

        elif isinstance(node, (Field, MetaRef)):
            return ctx.to_value(node)

        elif isinstance(node, TupleVariable):
            # exc("TupleVariable unsupported", "TupleVariable is not supported yet.", [err_source(source)])
            items = [self.lookup(ctx, item) for item in node._items]
            return tuple(items)

        elif isinstance(node, TableSchema):
            exc("Invalid value", "A schema cannot be used as a value outside of a `.new(..)` or `.select(..)` call.", [err_source(source)])
            return None

        elif is_primitive(node):
            type_ = self.to_type(Literal._get_type(node))
            return mLiteral(type_, node)

        elif isinstance(node, DerivedColumn):
            self.lookup(ctx, node._table)
            return ctx.to_value(node)

        elif isinstance(node, Literal):
            return mLiteral(type=self.to_type(node._type), value=node._value, source=source)

        elif isinstance(node, ModelEnum):
            return self.lookup(ctx, node._compile_lookup())

        elif isinstance(node, Alias):
            return self.lookup(ctx, node._source)

        elif isinstance(node, AsBool):
            # we need to create a match with two branches that returns true/false
            out = ctx.to_value(node)
            assert isinstance(out, Var)
            with ctx.subcontext():
                self.lookup(ctx, node._item)
                ctx.add_eq(out, mLiteral(bt.core.Boolean, True), source=node._source)
                true_branch = Logical(tuple(ctx.nodes), source=source)
            with ctx.subcontext():
                ctx.add_eq(out, mLiteral(bt.core.Boolean, False), source=node._source)
                false_branch = Logical(tuple(ctx.nodes), source=source)
            ctx.add(mMatch((true_branch, false_branch), source=source))
            return out

        #------------------------------------------------------
        # Invalid nodes
        #------------------------------------------------------

        elif isinstance(node, Distinct):
            ctx.err("Invalid distinct", "Distinct can only be used in `select(distinct(..))` or aggregates like `count(distinct(..))`.", [
                err_source(source)
            ])
            return None

        else:
            raise NotImplementedError(f"Lookup not implemented for node type: {type(node)}")

    #------------------------------------------------------
    # Update
    #------------------------------------------------------

    def update(self, ctx: Context, node: Any, root_source:SourcePos) -> mValue:
        source = node._source if not is_primitive(node) else root_source
        self.check_import(node)

        #------------------------------------------------------
        # Concept / Concept expressions (includes Data and Table)
        #------------------------------------------------------

        if isinstance(node, Concept):
            final_var = ctx.to_value(node)
            ctx.add_update(node, [final_var], source=source)
            return final_var

        elif isinstance(node, New):
            if node._identity_only:
                ctx.err("Invalid define", "Cannot define just an identity.", [
                    err_source(source),
                    "Did you mean to use `.new(..)`?"
                ])
                return
            return handle_concept_expression(self, ctx, node)

        #------------------------------------------------------
        # Relationship
        #------------------------------------------------------

        elif isinstance(node, Relationship):
            ctx.err("Invalid define", "Cannot define a relationship without values for its fields.", [
                err_source(root_source),
            ])

        #------------------------------------------------------
        # Expressions
        #------------------------------------------------------

        elif isinstance(node, Expression):
            if node._auto_filled or len(node._args) == 0:
                field_name = node._op._fields[-1].name if isinstance(node._op, Relationship) else node._op._name.lower()
                ctx.err("Invalid define", f"Define requires that all args be provided, but the value for the `{field_name}` field is missing.", [
                    err_source(source),
                ])
                return
            args = [self.lookup(ctx, arg) for arg in node._args]
            if node._op is CoreRelationships["="]:
                if isinstance(node._args[0], (Relationship, Chain)):
                    return self.update(ctx, node._args[0](node._args[1]), root_source=source)
                elif isinstance(node._args[1], (Relationship, Chain)):
                    return self.update(ctx, node._args[1](node._args[0]), root_source=source)
                else:
                    ctx.err("Invalid define", "Cannot set a non-relationship via `=` in a define.", [
                        err_source(source),
                    ])
            else:
                if isinstance(node._op, Concept) and is_core_concept(node._op):
                    return self.lookup(ctx, node)
                ctx.add_update(node._op, args, source=source)
                return args[-1]

        elif isinstance(node, Chain):
            # all fields have to be specified for a define, so a chain only works
            # if the relationship is unary
            if len(node._next._fields) > 1:
                ctx.err("Invalid define", "All fields must have a value when defining a relationship.", [
                    err_source(source),
                    f"You need something like [cyan]`.{node._next._short_name}(" + ", ".join([field.name for field in node._next._fields[1:]]) + ")`"
                ])
                return

            start = self.lookup(ctx, node._start)
            ctx.add_update(node._next, [start], source=source)
            return start

        #------------------------------------------------------
        # Invalids
        #------------------------------------------------------

        elif is_primitive(node):
            ctx.err("Invalid define", "Cannot define a primitive value.", [
                err_source(source)
            ])
            return

        elif isinstance(node, (FilterBy, Ref, FieldRef, Aggregate, Group, Not, Fragment, Match, Field, DerivedColumn, Literal, Alias, Distinct)):
            ctx.err("Invalid define", f"Cannot define a value of type `{type(node).__name__}`.", [
                err_source(node._source)
            ])
            return

        else:
            raise NotImplementedError(f"Lookup not implemented for node type: {type(node)}")



#------------------------------------------------------
# TODO
#------------------------------------------------------
"""
- vars shared between columns need to end up as column keys
- check sub-selects now that we went back to optional logicals
- roots for union/matches/fragments
- ref scheme mappings
- what do we want to do about Error/require errors?
- None columns in selects

x Concept
x DerivedColumn
x Field
x Relationship
x Reading
x Property
x Literal
x Chain
x Expression
x FieldRef
x Alias
x Match / Union
x Not
x Distinct
x nested Fragments
x New
x Concept expression with kwargs
x select
x where
x Data
x Table
x Group
x Aggregate
x defining an expression allows you to have one fewer var than you should
x define
x require
x distinct for aggregates
x don't do pop checks on primitives
x handle library imports
x nested news
x bool handling (as_bool, select(x > y))
x generate constraints for Property?
x wrong arg count in an update isn't providing an error
"""



# This script runs a QB program and captures the Snowflake requests made to the RAI Native App.
# The output is a JSON file containing information about the requests that can be used to run
# the same queries on a local RAI server.

import argparse
import os
import json
from contextlib import contextmanager

from v0.relationalai.clients.resources.snowflake import Resources as snowflake_api
from v0.relationalai.semantics.internal import internal
from typing import Dict, Optional

def main():
    parser = argparse.ArgumentParser(description="Extract LQP requests to run locally")
    parser.add_argument("--program", type=str, help="Path to the QB program to run", required=True)
    parser.add_argument("--outdir", type=str, help="Path to the output directory", required=True)

    args = parser.parse_args()
    program = args.program
    outdir = args.outdir

    if not os.path.isfile(program):
        raise ValueError(f"Program path {program} is not a file")
    if not os.path.isdir(outdir):
        raise ValueError(f"Output path {outdir} is not a directory")

    extract_rai_calls(program, outdir)

@contextmanager
def instrumented_exec_rai_app(captured_calls, call_counter):
    """Context manager to instrument _exec_rai_app method for capturing RAI calls."""
    # Store the original method
    original_exec_rai = snowflake_api._exec_rai_app

    def _instrumented_exec_rai_app(
        self,
        database: str,
        engine,
        raw_code: str,
        inputs: Dict,
        readonly=True,
        nowait_durable=False,
        request_headers=None,
        bypass_index=False,
        language: str = "rel",
        query_timeout_mins: Optional[int] = None,
    ):
        result = original_exec_rai(
            self,
            database=database,
            engine=engine,
            raw_code=raw_code,
            inputs=inputs,
            readonly=readonly,
            nowait_durable=nowait_durable,
            request_headers=request_headers,
            bypass_index=bypass_index,
            language=language,
            query_timeout_mins=query_timeout_mins,
        )

        call_counter[0] += 1
        exec_call_json = {
            "call_type": "raicode",
            "id": call_counter[0],
            "database": database,
            "inputs": inputs,
            "readonly": readonly,
            "nowait_durable": nowait_durable,
            "language": language,
            "timeout_mins": query_timeout_mins,
            "raw_code": raw_code,
        }
        captured_calls.append(exec_call_json)
        return result

    # Apply the monkey patch
    snowflake_api._exec_rai_app = _instrumented_exec_rai_app

    try:
        yield
    finally:
        # Restore the original method
        snowflake_api._exec_rai_app = original_exec_rai

def extract_rai_calls(qb_prog_path: str, outdir: str):
    """Extract RAI calls by instrumenting _exec_rai_app and executing the QB program."""
    exec_calls = []
    call_counter = [0]  # Use list for mutable reference in closure

    with instrumented_exec_rai_app(exec_calls, call_counter):
        with internal.with_overrides(**{
            'dry_run': False,
            'reasoner.rule.use_lqp': True,
        }):
            with open(qb_prog_path) as f:
                # Compile with the correct filename
                code = compile(f.read(), qb_prog_path, "exec")
                exec(code)

    # Write results to file
    final_json = json.dumps(exec_calls, default=str)
    out_path = os.path.join(outdir, "lqp_requests.json")
    with open(out_path, "w") as f:
        f.write(final_json)
    print(f"Wrote {len(exec_calls)} exec calls to {out_path}")

if __name__ == "__main__":
    main()



from __future__ import annotations
from contextlib import contextmanager
import difflib
import hashlib
from abc import ABC, abstractmethod
from pathlib import Path
from typing import cast, Type

from v0.relationalai.clients.types import ImportSource, ImportSourceFile
from v0.relationalai.clients.client import ResourcesBase
from v0.relationalai.dsl import Context, Graph, Instance, RelationRef, Vars, create_var
from v0.relationalai.loaders.types import LoadType, UnsupportedTypeError
from v0.relationalai.metamodel import Builtins
from v0.relationalai.rel_emitter import sanitize
from v0.relationalai.rel_utils import assert_no_problems, process_gnf_results
from v0.relationalai.std import rel

#-------------------------------------------------------------------------------
# Schema Mapping
#-------------------------------------------------------------------------------

TYPE_TO_REL_SCHEMA = {
    Builtins.String: "string",
    Builtins.Int: "int",
    Builtins.Number: "float",
    Builtins.Bool: "boolean",
    Builtins.Decimal: "decimal" # @TODO: See if this works or requires manually specifying size
}

# @NOTE: This only works so long as the mapping is bidirectionally unambiguous
REL_SCHEMA_TO_TYPE = {v: k for k, v in TYPE_TO_REL_SCHEMA.items()}

def rel_schema_to_type(rel_type: str):
    try:
        return REL_SCHEMA_TO_TYPE[rel_type]
    except KeyError:
        raise Exception(f"Invalid schema field type '{rel_type}'. Available options are ({', '.join(REL_SCHEMA_TO_TYPE.keys())})")

#-------------------------------------------------------------------------------
# Hashing
#-------------------------------------------------------------------------------

def compute_file_hash(file_path:str|Path, chunk_size=16384):
    """Digest the given file to a content-addressable hash."""
    sha256 = hashlib.sha256()
    with open(file_path, 'rb') as f:
        while chunk := f.read(chunk_size):
            sha256.update(chunk)
    return sha256.hexdigest()

def compute_str_hash(content: str):
    """Digest the given string to a hash."""
    return hashlib.sha256(content.encode()).hexdigest()

#-------------------------------------------------------------------------------
# Resource Proxy
#-------------------------------------------------------------------------------

class ResourceProxy:
    """Represents a (optimistically-assumed to be present) resource relation, such as a CSV."""
    def __init__(self, graph: Graph, name: str, id: Instance|None = None):
        if id is None:
            id = create_var()

        self._graph = graph
        self._rel = getattr(rel, sanitize(name)) # @FIXME: This isn't necessarily the same as the relation name, gota do something about that.
        self._refs: dict[str, RelationRef] = {}
        self._id = cast(Instance, id)

    def __getattr__(self, attr: str):
        ref = self._refs.get(attr, None)
        if ref is not None:
            return ref

        rel = getattr(self._rel, attr)
        attr_var = Vars(1)
        rel(self._id, attr_var)
        return attr_var

# This could instead be a subclass of Context, it behaves exactly as if it was
# but doing so happened to be larger and (seemingly?) more brittle due to all the
# kwargs.
@contextmanager
def read_resource_context(graph: Graph, name: str, **kwargs):
    ctx = Context(graph, **kwargs)
    with ctx:
        yield ResourceProxy(graph, name)


#-------------------------------------------------------------------------------
# Loader
#-------------------------------------------------------------------------------

class Loader(ABC):
    """Base class for objects that handle the particulars of mapping an import source to a resource relation."""
    type: str = "untyped"

    @abstractmethod
    def load(self, provider: ResourcesBase, model: str, source: ImportSource):
        """Load a snapshot of the given source data into the given model."""
        ...

    @classmethod
    def register(cls):
        assert cls.type != "untyped", f"{cls.__name__} does not provide a type property."
        assert cls.type not in Loader.type_to_loader or Loader.type_to_loader[cls.type] == cls, f"{cls.__name__} has the same type '{cls.type}' as another loader ({Loader.type_to_loader[cls.type].__name__})."
        Loader.type_to_loader[cls.type] = cls

    @classmethod
    def register_for_extensions(cls, *exts: str):
        cls.register()
        for ext in exts:
            assert ext not in Loader.ext_to_type or Loader.ext_to_type[ext] == cls.type, f"{cls.__name__} registered to handle '{ext}' but another loader ({Loader.type_to_loader[cls.type].__name__}) already does."
            Loader.ext_to_type[ext] = cls.type

    ext_to_type: dict[str, LoadType] = {}

    @classmethod
    def get_type_for(cls, source: ImportSourceFile):
        try:
            return Loader.ext_to_type[source.extension]
        except KeyError:
            raise UnsupportedTypeError(f"No loader type provided for extension: '{source.extension}'. You can manually specify the loader to use via the `type` kwarg", source.extension)

    type_to_loader: dict[LoadType, 'Type[Loader]'] = {}
    @classmethod
    def get_loader_for_type(cls, type: LoadType):
       try:
           return Loader.type_to_loader[type]
       except KeyError:
           raise UnsupportedTypeError(f"No loader provided for type: '{type}'", type)

#-------------------------------------------------------------------------------
# Import Utilities
#-------------------------------------------------------------------------------

def import_file(provider: ResourcesBase, model: str, source: ImportSourceFile, type: LoadType = "auto", **metadata):
    """Automatically load `source` into `model` using the most appropriate loader."""
    if type == "auto":
        type = Loader.get_type_for(source)

    return Loader.get_loader_for_type(type)().load(provider, model, source, **metadata)

def emit_delete_import(name: str) -> str:
    """Utility fn for emitting the rel to delete the specified import relation and all its metadata."""
    relation = sanitize(name)
    return f"""
    declare {relation}
    def delete[:{relation}]: {relation}
    def delete[:__resource, k]: ("{name}", __resource[k, "{name}"])
    """

#-------------------------------------------------------------------------------
# Resource Management
#-------------------------------------------------------------------------------

available_resources:dict[str, dict] = {}

def list_available_resources(provider: ResourcesBase, database: str, engine: str):
    res = provider.exec_raw(database, engine, """
    declare __resource
    def output {{__resource}}
    """)
    assert_no_problems(res)

    available_resources = process_gnf_results(res.results, "name")
    return available_resources

class MissingResourceError(Exception):
    def __init__(self, name: str):
        similar_names = difflib.get_close_matches(name, available_resources.keys(), n=3)
        message = f"No available resource named '{name}'"
        if similar_names:
            message += ". Maybe you want one of these?"
            for name in similar_names:
                message += f"\n- {name}"
        super().__init__(message)



from __future__ import annotations

from v0.relationalai.semantics.metamodel import ir
from v0.relationalai.semantics.metamodel.compiler import Pass
from v0.relationalai.semantics.metamodel.visitor import Rewriter
from v0.relationalai.semantics.metamodel.util import OrderedSet
from v0.relationalai.semantics.metamodel import helpers, factory as f, types, builtins
from typing import Optional, Any, Iterable, Tuple
from collections import defaultdict

"""
Given an Output with a group of keys (some of them potentially null),
* extract the lookups that bind (transitively) all the keys
* extract the lookups that bind (transitively) properties of the keys
* generate all the valid combinations of keys being present or not
  * first all keys are present,
  * then we remove one key at a time,
  * then we remove two keys at a time,and so on.
  * the last combination is when all the *nullable* keys are missing.
* for each combination:
  * create a compound (hash) key
  * create a Logical that:
      * contains all the relevant lookups for keys and properties
      * contains negated lookups for null keys
      * outputs using the compound key

E.g., we go from

Logical
    Foo(foo)
    rel1(foo, x)
    Logical ^[v1=None]
        rel2(foo, v1)
    Logical ^[v2=None, k2=None]
        rel3(foo, k2)
        rel4(k2, v2)
    Logical ^[v3=None, k3=None]
        rel5(foo, y)
        rel6(y, k3)
        rel7(k3, v3)
    output[foo, k2, k3](v1, v2, v3)

to

Logical
    Logical
        Logical
            Foo(foo)
            rel1(foo, x)
            Logical ^[k2=None, k3=None]
                rel3(foo, k2)
                rel5(foo, y)
                rel6(y, k3)
            Logical ^[v1=None]
                rel2(foo, v1)
            Logical ^[v2=None, k2=None]
                rel4(k2, v2)
            Logical ^[v3=None, k3=None]
                rel7(k3, v3)
            construct(Hash, "Foo", foo, "Concept2", k2, "Concept3", k3, compound_key)
            output[compound_key](v1, v2, v3)
        Logical
            Foo(foo)
            rel1(foo, x)
            Logical ^[k2=None, k3=None]
                rel5(foo, y)
                rel6(y, k3)
            Not
                Logical
                    rel3(foo, k2)
            Logical ^[v1=None]
                rel2(foo, v1)
            Logical ^[v3=None, k3=None]
                rel7(k3, v3)
            construct(Hash, "Foo", foo, "Concept3", k3, compound_key)
            output[compound_key](v1, None, v3)
        Logical
            Foo(foo)
            rel1(foo, x)
            Logical ^[k2=None, k3=None]
                rel3(foo, k2)
                rel5(foo, y)
            Not
                Logical
                    rel6(y, k3)
            Logical ^[v1=None]
                rel2(foo, v1)
            Logical ^[v2=None, k2=None]
                rel4(k2, v2)
            construct(Hash, "Foo", foo, "Concept2", k2, compound_key)
            output[compound_key](v1, v2, None)
        Logical
            Foo(foo)
            rel1(foo, x)
            Logical ^[k2=None, k3=None]
                rel5(foo, y)
            Not
                Logical
                    rel3(foo, k2)
                    rel6(y, k3)
            Logical ^[v1=None]
                rel2(foo, v1)
            construct(Hash, "Foo", foo, compound_key)
            output[compound_key](v1, None, None)
"""
class ExtractKeys(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        return ExtractKeysRewriter().walk(model)

"""
* First, figure out all tasks that are common for all alternative logicals that will be
  generated
* Second, generate the appropriate negated tasks for when some keys are missing
* Third, add tasks that are needed for the final output columns
  (e.g., lookup name if we select name)
* Lastly, create the approprite Output task, with some of the columns being None if they are
  missing (None will be filtered out in a later step -- we just need the column number to be
  the same here).
"""
class ExtractKeysRewriter(Rewriter):
    def handle_logical(self, node: ir.Logical, parent: ir.Node, ctx:Optional[Any]=None) -> ir.Logical:
        outputs = [x for x in node.body if isinstance(x, ir.Output) and x.keys]
        # We are not in a logical with an output at this level.
        if not outputs:
            new_body = self.walk_list(node.body, node)
            return node if new_body is node.body else f.logical(new_body, node.hoisted)
        assert len(outputs) == 1, "multiple outputs with keys in a logical"
        output = outputs[0]
        assert output.keys
        output_keys = output.keys

        all_vars = OrderedSet.from_iterable(output_keys)
        all_vars.update(helpers.output_vars(output.aliases))

        info = self.preprocess_logical(node, output_keys)
        top_level_tasks, var_to_default, nullable_keys, non_nullable_keys = info
        # we only need to transform the logical if there are nullable keys
        if not nullable_keys:
            return node

        # Flatten the contents of (direct) logicals that comprise the body of this output
        flat_body = OrderedSet[ir.Task]()
        for task in node.body:
            if isinstance(task, ir.Logical):
                if any(isinstance(t, (ir.Aggregate, ir.Rank)) for t in task.body):
                    flat_body.add(task)
                else:
                    flat_body.update(task.body)
            elif isinstance(task, ir.Output):
                continue
            else:
                flat_body.add(task)

        # Add to all_vars the vars of lookups that were flattened above (from some logical)
        for task in flat_body:
            if isinstance(task, ir.Lookup):
                all_vars.update(helpers.vars(task.args))

        partitions, deps = self.partition_tasks(flat_body, all_vars)

        # Compute all valid key combinations (keys that are not null)
        combinations = self.key_combinations(nullable_keys, deps, 0, non_nullable_keys.get_list())
        # there is no need to transform if there is only a single combination
        if len(combinations) == 1:
            return node

        ####################################################################################

        outer_body: list[ir.Task] = []
        annos = list(output.annotations)
        annos.append(f.annotation(builtins.output_keys, tuple(output_keys)))
        # Create a compound key that will be used in place of the original keys.
        compound_key = f.var("compound_key", types.Hash)

        for key_combination in combinations:
            missing_keys = OrderedSet.from_iterable(output_keys)
            missing_keys.difference_update(key_combination)

            # top-level tasks are always present in every clone
            body = OrderedSet.from_iterable(top_level_tasks)
            # add tasks that are specific to the keys present in the current combination
            for key in key_combination:
                assert key in partitions
                body.update(partitions[key])

            # vars used in a positive context
            positive_vars = OrderedSet[ir.Var]()
            for task in body:
                if isinstance(task, ir.Lookup):
                    positive_vars.update(helpers.vars(task.args))

            # handle the construct node in each clone
            values: list[ir.Value] = [compound_key.type]
            for key in output_keys:
                assert isinstance(key.type, ir.ScalarType)
                values.append(ir.Literal(types.String, key.type.name))
                if key in key_combination:
                    values.append(key)
            body.add(ir.Construct(None, tuple(values), compound_key, OrderedSet().frozen()))

            # find variables used only inside the negated context
            negative_vars = OrderedSet[ir.Var]()
            for key in missing_keys:
                negative_partition = partitions[key]
                for task in negative_partition:
                    if isinstance(task, ir.Lookup):
                        args = helpers.vars(task.args)
                        for arg in args:
                            if arg not in positive_vars:
                                negative_vars.add(arg)

            out_vars = helpers.output_vars(output.aliases)
            # output variables that depend on a missing key. They have to be skipped
            missing_out_vars = OrderedSet[ir.Var]()
            # output variables that depend on a negated variable that's not a missing key
            # they have to be handled in a more complicated manner
            problematic_out_vars = OrderedSet[ir.Var]()
            for out_var in out_vars:
                out_deps = deps[out_var]
                if out_var in missing_keys:
                    missing_out_vars.add(out_var)
                elif any(x in missing_keys for x in out_deps):
                    missing_out_vars.add(out_var)
                elif any(x in negative_vars for x in out_deps):
                    problematic_out_vars.add(out_var)

            if problematic_out_vars:
                assert len(problematic_out_vars) == 1
                exclude_vars = deps[problematic_out_vars[0]] & negative_vars
                has_problematic_var = True
            else:
                exclude_vars = out_vars
                has_problematic_var = False

            self.negate_missing_keys(body, missing_keys, var_to_default, partitions, deps,
                                     out_vars, exclude_vars, negative_vars, has_problematic_var)

            new_output_aliases = []
            for alias, out_value in output.aliases:
                if not isinstance(out_value, ir.Var):
                    new_out_value = out_value
                else:
                    new_out_value = None if out_value in missing_out_vars else out_value
                new_output_aliases.append((alias, new_out_value))
            body.add(f.output(new_output_aliases, [compound_key], annos=annos))

            # Create the final logical for this combination
            outer_body.append(f.logical(tuple(body), []))

        return f.logical(tuple(outer_body), [])

    def noop_logical(self, node: ir.Logical) -> bool:
        # logicals that don't hoist variables are essentially filters like lookups
        if not node.hoisted:
            return True
        if len(node.body) != 1:
            return False
        inner = node.body[0]
        if not isinstance(inner, (ir.Match, ir.Union)):
            return False
        outer_vars = helpers.hoisted_vars(node.hoisted)
        inner_vars = helpers.hoisted_vars(inner.hoisted)
        for v in outer_vars:
            if v not in inner_vars:
                return False
        # all vars hoisted by the outer logical, are also
        # hoisted by the inner Match/Union
        return True

    # compute inital information that's needed for later steps. E.g., what's nullable or
    # not, do some output columns have a default value, etc.
    def preprocess_logical(self, node: ir.Logical, output_keys: Iterable[ir.Var]):
        top_level_tasks = OrderedSet()
        non_nullable_vars = OrderedSet()
        nullable_vars = OrderedSet()
        var_to_default: dict[ir.Var, ir.Default] = {}

        for task in node.body:
            # Top-level lookups use only non-nullable vars
            if isinstance(task, ir.Lookup):
                vars = helpers.vars(task.args)
                non_nullable_vars.update(vars)
                top_level_tasks.add(task)
            elif isinstance(task, ir.Logical):
                if self.noop_logical(task):
                    top_level_tasks.add(task)
                    non_nullable_vars.update(helpers.hoisted_vars(task.hoisted))
                    continue

                for h in task.hoisted:
                    # Hoisted vars without a default are not nullable
                    if isinstance(h, ir.Var):
                        non_nullable_vars.add(h)
                    elif isinstance(h, ir.Default):
                        var_to_default[h.var] = h
                        # Hoisted vars with a non-None default are not nullable
                        if h.value is not None:
                            non_nullable_vars.add(h.var)
                        else:
                            nullable_vars.add(h.var)

            # Variables appearing in an aggregate's group-by are not nullable
            elif isinstance(task, ir.Aggregate):
                top_level_tasks.add(task)
                for v in task.group:
                    non_nullable_vars.add(v)

            elif isinstance(task, ir.Construct):
                top_level_tasks.add(task)
                non_nullable_vars.add(task.id_var)
            elif isinstance(task, ir.Data):
                top_level_tasks.add(task)
                non_nullable_vars.update(task.vars)
            # TODO: should Union and Match be used for hoisted vars?
            elif isinstance(task, (ir.Not, ir.Match, ir.Union)):
                top_level_tasks.add(task)

            else:
                if not isinstance(task, ir.Output):
                    raise ValueError(f"Unexpected task type: {type(task)}")

        # Any variable appearing in both sets is non-nullable
        nullable_vars = nullable_vars - non_nullable_vars

        non_nullable_keys = OrderedSet.from_iterable(output_keys)
        non_nullable_keys.difference_update(nullable_vars)
        nullable_keys = OrderedSet.from_iterable(output_keys)
        nullable_keys.difference_update(non_nullable_keys)

        return top_level_tasks, var_to_default, nullable_keys, non_nullable_keys

    # given a set of variables, compute the tasks that each variable is using and also
    # other variables needed for this variable to bind correctly
    def partition_tasks(self, tasks:Iterable[ir.Task], vars:Iterable[ir.Var]):
        partitions:dict[ir.Var, OrderedSet[ir.Task]] = defaultdict(OrderedSet)
        dependencies:dict[ir.Var, OrderedSet[ir.Var]] = defaultdict(OrderedSet)

        def dfs_collect_deps(task, deps):
            if isinstance(task, ir.Lookup):
                args = helpers.vars(task.args)
                for i, v in enumerate(args):
                    # v depends on all previous vars
                    for j in range(i):
                        deps[v].add(args[j])
                    # for ternary+ lookups, a var also depends on the next vars
                    if i > 0 and len(args) >= 3:
                        for j in range(i+1, len(args)):
                            deps[v].add(args[j])
            elif isinstance(task, ir.Construct):
                vars = helpers.vars(task.values)
                for val_var in vars:
                    deps[task.id_var].add(val_var)
            elif isinstance(task, ir.Logical):
                for child in task.body:
                    dfs_collect_deps(child, deps)
            elif isinstance(task, (ir.Match, ir.Union)):
                for child in task.tasks:
                    dfs_collect_deps(child, deps)

        for task in tasks:
            dfs_collect_deps(task, dependencies)

        def dfs_transitive_deps(var, visited):
            for dep_var in dependencies[var]:
                if dep_var not in visited:
                    visited.add(dep_var)
                    dfs_transitive_deps(dep_var, visited)

        transitive_deps = defaultdict(OrderedSet)
        for var in list(dependencies.keys()):
            visited = OrderedSet()
            dfs_transitive_deps(var, visited)
            transitive_deps[var] = visited
        dependencies = transitive_deps

        for var in vars:
            extended_vars = OrderedSet[ir.Var]()
            extended_vars.add(var)

            there_is_progress = True
            while there_is_progress:
                there_is_progress = False
                for task in tasks:
                    if task in partitions[var]:
                        continue

                    if isinstance(task, (ir.Logical, ir.Match, ir.Union)):
                        hoisted = helpers.hoisted_vars(task.hoisted)
                        if var in hoisted:
                            partitions[var].add(task)
                            there_is_progress = True
                    elif isinstance(task, ir.Construct):
                        if task.id_var == var:
                            partitions[var].add(task)
                            there_is_progress = True
                    elif isinstance(task, ir.Lookup):
                        args = helpers.vars(task.args)
                        if len(args) == 1 and args[0] in extended_vars:
                            partitions[var].add(task)
                            there_is_progress = True
                        # NOTE: heuristics to have dot_joins work
                        elif len(args) >= 3 and args[-2] in extended_vars:
                            partitions[var].add(task)
                            extended_vars.add(args[-1])
                            there_is_progress = True
                        elif len(args) > 1 and args[-1] in extended_vars:
                            partitions[var].add(task)
                            for arg in args[:-1]:
                                extended_vars.add(arg)
                            there_is_progress = True
                    else:
                        assert False, f"invalid node kind {type(task)}"

        return partitions, dependencies

    # Generate all the valid combinations of non-nullable keys and nullable keys.
    def key_combinations(self, nullable_keys: OrderedSet[ir.Var], key_deps, idx: int, non_null_keys: list[ir.Var]) -> OrderedSet[Tuple[ir.Var]]:
        if idx < len(nullable_keys):
            key = nullable_keys[idx]
            set1 = self.key_combinations(nullable_keys, key_deps, idx + 1, non_null_keys + [key])
            set2 = self.key_combinations(nullable_keys, key_deps, idx + 1, non_null_keys)
            set1.update(set2)
            return set1
        else:
            final_keys = []
            for k in non_null_keys:
                # If a key depends on other keys, all of them should be present in this combination.
                # If some dependency is not present, ignore the current key.
                deps = key_deps.get(k)
                if deps and any(dk in nullable_keys and dk not in non_null_keys for dk in deps):
                    continue
                final_keys.append(k)
            return OrderedSet.from_iterable([tuple(final_keys)])

    def negate_missing_keys(self, body, missing_keys, var_to_default, partitions, deps,
                            out_vars, exclude_vars, negative_vars, has_problematic_var:bool):
        # for keys that are not present in the current combination
        # we have to include their tasks negated
        negated_tasks = OrderedSet[ir.Task]()
        positive_tasks = defaultdict(OrderedSet)

        for key in missing_keys:
            negative_body = OrderedSet[ir.Task]()
            negative_partition = partitions[key]
            for task in negative_partition:
                # task is already present positively at the top-level context
                if task in body:
                    continue
                if not isinstance(task, ir.Lookup):
                    negative_body.add(task)
                    continue

                args = helpers.vars(task.args)
                if args[-1] in exclude_vars and args[-1] not in missing_keys:
                    positive_tasks[args[-1]].add(task)
                else:
                    negative_body.add(task)

            negated_tasks.update(negative_body)
            if len(negative_body) > 1:
                body.add(f.not_(f.logical(tuple(negative_body), [])))
            elif len(negative_body) == 1:
                body.add(f.not_(negative_body[0]))

        if positive_tasks:
            for _, d in positive_tasks.items():
                body.update(d)

        for out_var in out_vars:
            out_deps = deps[out_var]
            if has_problematic_var and any(x in missing_keys for x in out_deps):
                continue
            elif not has_problematic_var and any(x in missing_keys or x in negative_vars for x in out_deps):
                continue

            default = var_to_default.get(out_var)
            partition = OrderedSet.from_iterable(partitions[out_var])
            partition.difference_update(body)
            property_body = OrderedSet[ir.Task]()
            if not has_problematic_var:
                for task in partition:
                    if isinstance(task, ir.Lookup):
                        vars = helpers.vars(task.args)
                        if any(x in negative_vars for x in vars):
                            continue
                    property_body.add(task)
            else:
                property_body.update(partition)
            if property_body:
                body.add(f.logical(tuple(property_body), [default] if default else []))
